import nbformat

import jupytext
from jupytext.compare import compare_notebooks


def test_identity_source_write_read(ipynb_py_file):
    """Test that writing the notebook with jupytext, and read again,
    is the same as removing outputs"""

    with open(ipynb_py_file) as fp:
        nb1 = nbformat.read(fp, as_version=4)

    py = jupytext.writes(nb1, "py")
    nb2 = jupytext.reads(py, "py")

    compare_notebooks(nb2, nb1)



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

"""
This notebook contains outputs of many different types: text, HTML, plots and errors.
"""

###############################################################################
# # Text outputs
#
# Using `print`, `sys.stdout` and `sys.stderr`

import sys
print('using print')
sys.stdout.write('using sys.stdout.write')
sys.stderr.write('using sys.stderr.write')

""
import logging
logging.debug('Debug')
logging.info('Info')
logging.warning('Warning')
logging.error('Error')

###############################################################################
# # HTML outputs
#
# Using `pandas`. Here we find two representations: both text and HTML.

import pandas as pd
pd.DataFrame([4])

""
from IPython.display import display
display(pd.DataFrame([5]))
display(pd.DataFrame([6]))

###############################################################################
# # Images

# %matplotlib inline

""
# First plot
from matplotlib import pyplot as plt
import numpy as np
w, h = 3, 3
data = np.zeros((h, w, 3), dtype=np.uint8)
data[0,:] = [0,255,0]
data[1,:] = [0,0,255]
data[2,:] = [0,255,0]
data[1:3,1:3] = [255, 0, 0]
plt.imshow(data)
plt.axis('off')
plt.show()
# Second plot
data[1:3,1:3] = [255, 255, 0]
plt.imshow(data)
plt.axis('off')
plt.show()

###############################################################################
# # Errors

undefined_variable



import jupytext
from jupytext.compare import compare


def test_identity_source_write_read(py_file):
    with open(py_file) as fp:
        py = fp.read()

    nb = jupytext.reads(py, "py")
    py2 = jupytext.writes(nb, "py")

    compare(py2, py)



import jupytext

from .combine import combine_inputs_with_outputs
from .compare import compare
from .formats import check_file_version, long_form_multiple_formats
from .paired_paths import find_base_path_and_format, full_path
from .pairs import PairedFilesDiffer


async def read_pair(inputs, outputs, read_one_file, must_match=False):
    """Read a notebook given its inputs and outputs path and formats"""
    if not outputs.path or outputs.path == inputs.path:
        return await read_one_file(inputs.path, inputs.fmt)

    notebook = await read_one_file(inputs.path, inputs.fmt)
    check_file_version(notebook, inputs.path, outputs.path)

    notebook_with_outputs = await read_one_file(outputs.path, outputs.fmt)

    if must_match:
        in_text = jupytext.writes(notebook, inputs.fmt)
        out_text = jupytext.writes(notebook_with_outputs, inputs.fmt)
        diff = compare(out_text, in_text, outputs.path, inputs.path, return_diff=True)
        if diff:
            raise PairedFilesDiffer(diff)

    notebook = combine_inputs_with_outputs(
        notebook, notebook_with_outputs, fmt=inputs.fmt
    )

    return notebook


async def write_pair(path, formats, write_one_file):
    """
    Call the function 'write_one_file' on each of the paired path/formats
    """
    formats = long_form_multiple_formats(formats)
    base, _ = find_base_path_and_format(path, formats)

    # Save as ipynb first
    return_value = None
    value = None
    ipynb_changed = False
    for fmt in formats[::-1]:
        if fmt["extension"] != ".ipynb":
            continue

        alt_path = full_path(base, fmt)
        value = await write_one_file(alt_path, fmt)
        ipynb_changed = ipynb_changed or value.get("modified", False)
        if alt_path == path:
            return_value = value

    # And then to the other formats, in reverse order so that
    # the first format is the most recent
    for fmt in formats[::-1]:
        if fmt["extension"] == ".ipynb":
            continue

        alt_path = full_path(base, fmt)
        if ipynb_changed:
            value = await write_one_file(alt_path, fmt, force_update_timestamp=True)
        else:
            # in the contents manager the write_one_file always writes anyway
            value = await write_one_file(alt_path, fmt)
        if alt_path == path:
            return_value = value

    # Update modified timestamp to match that of the pair #207
    if isinstance(return_value, dict) and "last_modified" in return_value:
        return_value["last_modified"] = value["last_modified"]

    return return_value



from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_raw_cell


def test_cell_id_is_not_random():
    assert new_code_cell().id == "cell-1"
    assert new_code_cell().id == "cell-2"
    assert new_markdown_cell().id == "cell-3"
    assert new_raw_cell().id == "cell-4"



import json

import pytest

from jupytext_config.labconfig import LabConfig


@pytest.fixture()
def sample_viewer_config():
    return {
        "@jupyterlab/docmanager-extension:plugin": {
            "defaultViewers": {
                "markdown": "Jupytext Notebook",
                "myst": "Jupytext Notebook",
                "r-markdown": "Jupytext Notebook",
                "quarto": "Jupytext Notebook",
                "julia": "Jupytext Notebook",
                "python": "Jupytext Notebook",
                "r": "Jupytext Notebook",
            }
        }
    }


@pytest.fixture()
def sample_empty_viewer_config():
    return {"@jupyterlab/docmanager-extension:plugin": {"defaultViewers": {}}}


@pytest.fixture()
def settings_file(tmp_path):
    return tmp_path / "default_setting_overrides.json"


def test_read_config(settings_file, sample_viewer_config):
    (settings_file).write_text(json.dumps(sample_viewer_config))
    labconfig = LabConfig(settings_file=settings_file).read()
    assert labconfig.config == sample_viewer_config


def test_set_unset_default_viewers(
    settings_file, sample_viewer_config, sample_empty_viewer_config
):
    labconfig = LabConfig(settings_file=settings_file)
    labconfig.set_default_viewers()
    assert labconfig.config == sample_viewer_config
    labconfig.unset_default_viewers()
    assert labconfig.config == sample_empty_viewer_config


def test_write_config(settings_file, sample_viewer_config):
    labconfig = LabConfig(settings_file=settings_file)
    labconfig.set_default_viewers()
    labconfig.write()
    assert json.loads(settings_file.read_text()) == sample_viewer_config



import pytest

from jupytext.cli import jupytext


def test_run_pytest_ok(
    tmpdir,
    text="""This is a notebook with a test that is supposed to pass
```python
def test_ok():
    assert True
```
""",
):
    tmp_md = str(tmpdir.join("notebook.md"))
    with open(tmp_md, "w") as fp:
        fp.write(text)
    jupytext([tmp_md, "--check", "pytest"])


def test_run_pytest_fail(
    tmpdir,
    capsys,
    text="""This is a notebook with a test that should not pass
```python
def test_fail():
    assert False
```
""",
):
    tmp_md = str(tmpdir.join("notebook.md"))
    with open(tmp_md, "w") as fp:
        fp.write(text)
    with pytest.raises(SystemExit):
        jupytext([tmp_md, "--check", "pytest"])



import os

import pytest
from jupyter_server.utils import ensure_async
from nbformat.v4.nbbase import new_notebook
from nbformat.validator import NotebookValidationError
from tornado.web import HTTPError

import jupytext
from jupytext.compare import compare_notebooks, notebook_model

pytestmark = pytest.mark.asyncio


async def test_rmd_is_ok(ipynb_file, tmpdir, cm):
    nb = jupytext.read(ipynb_file)
    tmp_ipynb = "notebook.ipynb"
    tmp_rmd = "notebook.Rmd"

    nb.metadata.setdefault("jupytext", {})["formats"] = "ipynb,Rmd"

    cm.root_dir = str(tmpdir)

    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    nb2 = jupytext.read(str(tmpdir.join(tmp_rmd)))

    compare_notebooks(nb2, nb, "Rmd")


async def test_ipynb_is_ok(rmd_file, tmpdir, cm):
    nb = jupytext.read(rmd_file)
    tmp_ipynb = "notebook.ipynb"
    tmp_rmd = "notebook.Rmd"

    cm.root_dir = str(tmpdir)
    cm.formats = "ipynb,Rmd"

    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_rmd))

    nb2 = jupytext.read(str(tmpdir.join(tmp_ipynb)))
    compare_notebooks(nb2, nb)


async def test_all_files_created(ipynb_py_file, tmpdir, cm):
    nb = jupytext.read(ipynb_py_file)
    tmp_ipynb = "notebook.ipynb"
    tmp_rmd = "notebook.Rmd"
    tmp_py = "notebook.py"
    nb.metadata["jupytext"] = {"formats": "ipynb,Rmd,py"}

    cm.root_dir = str(tmpdir)

    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    nb2 = jupytext.read(str(tmpdir.join(tmp_py)))
    compare_notebooks(nb2, nb)

    nb3 = jupytext.read(str(tmpdir.join(tmp_rmd)))
    compare_notebooks(nb3, nb, "Rmd")


async def test_no_files_created_on_no_format(tmpdir, cm):
    tmp_ipynb = "notebook.ipynb"
    tmp_rmd = "notebook.Rmd"
    tmp_py = "notebook.py"

    cm.root_dir = str(tmpdir)
    cm.formats = ""

    await ensure_async(
        cm.save(
            model=notebook_model(new_notebook(nbformat=4, metadata=dict())),
            path=tmp_ipynb,
        )
    )

    assert not os.path.isfile(str(tmpdir.join(tmp_py)))
    assert not os.path.isfile(str(tmpdir.join(tmp_rmd)))


async def test_raise_on_wrong_format(tmpdir, cm):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))

    cm.root_dir = str(tmpdir)

    with pytest.raises(HTTPError):
        await ensure_async(
            cm.save(
                path=tmp_ipynb,
                model=dict(
                    type="notebook",
                    content=new_notebook(
                        nbformat=4, metadata=dict(jupytext_formats=[".doc"])
                    ),
                ),
            )
        )


async def test_no_rmd_on_not_notebook(tmpdir, cm):
    tmp_ipynb = "notebook.ipynb"
    tmp_rmd = "notebook.Rmd"

    cm.root_dir = str(tmpdir)
    cm.formats = "ipynb,Rmd"

    with pytest.raises(HTTPError):
        await ensure_async(
            cm.save(
                model=dict(type="not notebook", content=new_notebook()), path=tmp_ipynb
            )
        )
    assert not os.path.isfile(str(tmpdir.join(tmp_rmd)))


async def test_no_rmd_on_not_v4(tmpdir, cm):
    tmp_ipynb = "notebook.ipynb"
    tmp_rmd = "notebook.Rmd"

    cm.root_dir = str(tmpdir)
    cm.formats = "ipynb,Rmd"

    with pytest.raises(NotebookValidationError):
        await ensure_async(
            cm.save(model=notebook_model(new_notebook(nbformat=3)), path=tmp_rmd)
        )

    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))



# This script does the "same" job as `jupyter labextension develop --overwrite .`
#
# Seems like we cannot use the upstream script on our repo organization.
# If we want to use the upstream script we need to have a setup.py or pyproject.toml
# in jupyterlab/jupyterlab-jupytext folder with the name of the package.
#
# We are only interested in making a symlink to the
# sys.prefix/share/jupyter/labextensions folder for development. We should be able
# to do it in simple script
#

import os

import jupyterlab_jupytext
from jupyterlab.federated_labextensions import build_labextension, develop_labextension


def main():
    """Create symlink in sys.prefix based on name of extension"""

    labexts = jupyterlab_jupytext._jupyter_labextension_paths()
    base_path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        "jupyterlab_jupytext",
    )

    for labext in labexts:
        src = os.path.join(base_path, labext["src"])
        dest = labext["dest"]
        print(f"Installing {src} -> {dest}")

        if not os.path.exists(src):
            build_labextension(base_path)

        full_dest = develop_labextension(
            src,
            overwrite=True,
            symlink=True,
            user=False,
            sys_prefix=True,
            labextensions_dir="",
            destination=dest,
        )
        print(f"Creating symlink at {full_dest}")


if __name__ == "__main__":
    main()



import pytest
from nbformat.notebooknode import NotebookNode

import jupytext
from jupytext.formats import NOTEBOOK_EXTENSIONS
from jupytext.myst import is_myst_available, myst_extensions
from jupytext.quarto import is_quarto_available


@pytest.mark.parametrize("ext", sorted(set(NOTEBOOK_EXTENSIONS) - {".ipynb"}))
def test_read_empty_text_notebook(ext, tmp_path):
    if ext == ".qmd" and not is_quarto_available(min_version="0.2.0"):
        pytest.skip("quarto is not available")
    if ext in myst_extensions(no_md=True) and not is_myst_available():
        pytest.skip("MyST is not available")

    empty_nb = (tmp_path / "notebook").with_suffix(ext)
    empty_nb.touch()

    nb = jupytext.read(empty_nb)

    assert isinstance(nb, NotebookNode)
    assert not nb.cells



import pytest
from jupyter_server.utils import ensure_async


@pytest.mark.asyncio
async def test_remove_encoding_907(tmp_path, python_notebook, cm):
    # Pair all notebooks to py:percent files
    (tmp_path / "jupytext.toml").write_text('formats="ipynb,py:percent"')

    # Create a contents manager
    cm.root_dir = str(tmp_path)

    # Save the notebook in Jupyter
    await ensure_async(
        cm.save(dict(type="notebook", content=python_notebook), path="nb.ipynb")
    )

    # No encoding is present in the py file
    py = (tmp_path / "nb.py").read_text()
    assert "coding" not in py

    # Add the encoding line
    py = "# -*- coding: utf-8 -*-\n" + py
    (tmp_path / "nb.py").write_text(py)

    # Reload the notebook
    nb = (await ensure_async(cm.get("nb.ipynb")))["content"]
    assert "encoding" in nb.metadata["jupytext"]

    # Save the notebook
    await ensure_async(cm.save(dict(type="notebook", content=nb), path="nb.ipynb"))

    # The encoding is still present in the py file
    py = (tmp_path / "nb.py").read_text()
    assert py.startswith("# -*- coding: utf-8 -*-")

    # Remove the encoding (mock ipyupgrade)
    py = "\n".join(py.splitlines()[1:])
    (tmp_path / "nb.py").write_text(py)

    # Reload the notebook - the encoding is not there anymore
    nb = (await ensure_async(cm.get("nb.ipynb")))["content"]
    assert "encoding" not in nb.metadata["jupytext"]

    # Save the notebook - the encoding is not there anymore
    py = (tmp_path / "nb.py").read_text()
    assert "coding" not in py



import jupytext
from jupytext.cell_metadata import parse_key_equal_value
from jupytext.compare import compare


def test_parse_metadata():
    assert parse_key_equal_value("--key value --key-2 .\\a\\b.cs") == {
        "incorrectly_encoded_metadata": "--key value --key-2 .\\a\\b.cs"
    }


def test_parse_double_hyphen_metadata():
    assert parse_key_equal_value("--key1 value1 --key2 value2") == {
        "incorrectly_encoded_metadata": "--key1 value1 --key2 value2"
    }


def test_read_dotnet_try_markdown(
    md="""This is a dotnet/try Markdown file, inspired
from this [post](https://devblogs.microsoft.com/dotnet/creating-interactive-net-documentation/)

``` cs --region methods --source-file .\\myapp\\Program.cs --project .\\myapp\\myapp.csproj
var name ="Rain";
Console.WriteLine($"Hello {name.ToUpper()}!");
```
""",
):
    # Read the notebook
    nb = jupytext.reads(md, fmt=".md")
    assert nb.metadata["jupytext"]["main_language"] == "csharp"
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[1].cell_type == "code"
    assert (
        nb.cells[1].source
        == """var name ="Rain";
Console.WriteLine($"Hello {name.ToUpper()}!");"""
    )
    compare(
        nb.cells[1].metadata,
        {
            "language": "cs",
            "incorrectly_encoded_metadata": "--region methods --source-file .\\myapp\\Program.cs --project .\\myapp\\myapp.csproj",  # noqa: E501
        },
    )

    # Round trip to Markdown
    md2 = jupytext.writes(nb, "md")
    compare(md2, md.replace("``` cs", "```cs"))



import pytest
from nbformat.v4.nbbase import new_code_cell, new_notebook

from jupytext import write
from jupytext.cli import jupytext


@pytest.fixture
def non_black_notebook(python_notebook):
    return new_notebook(metadata=python_notebook.metadata, cells=[new_code_cell("1+1")])


@pytest.mark.requires_black
def test_check_notebooks_left_or_right_black(python_notebook, tmpdir, cwd_tmpdir):
    write(python_notebook, str(tmpdir / "nb1.ipynb"))
    write(python_notebook, str(tmpdir / "nb2.ipynb"))

    jupytext(["*.ipynb", "--check", "black --check {}"])
    jupytext(["--check", "black --check {}", "*.ipynb"])


@pytest.mark.requires_black
def test_check_notebooks_left_or_right_not_black(
    non_black_notebook, tmpdir, cwd_tmpdir
):
    write(non_black_notebook, str(tmpdir / "nb1.ipynb"))
    write(non_black_notebook, str(tmpdir / "nb2.ipynb"))

    with pytest.raises(SystemExit):
        jupytext(["*.ipynb", "--check", "black --check {}"])

    with pytest.raises(SystemExit):
        jupytext(["--check", "black --check {}", "*.ipynb"])



"""Read notebook cells from their text representation"""

import re
from collections import defaultdict
from copy import copy
from itertools import count

from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_raw_cell

from .doxygen import doxygen_to_markdown
from .languages import _SCRIPT_EXTENSIONS

# Sphinx Gallery is an optional dependency. And we intercept the SyntaxError for #301
try:
    from sphinx_gallery.notebook import rst2md
except (ImportError, SyntaxError):  # pragma: no cover
    rst2md = None

from .cell_metadata import (
    is_active,
    is_json_metadata,
    rmd_options_to_metadata,
    text_to_metadata,
)
from .languages import _JUPYTER_LANGUAGES_LOWER_AND_UPPER
from .magics import is_magic, need_explicit_marker, uncomment_magic, unescape_code_start
from .pep8 import pep8_lines_between_cells
from .stringparser import StringParser

_BLANK_LINE = re.compile(r"^\s*$")
_PY_INDENTED = re.compile(r"^\s")


def uncomment(lines, prefix="#", suffix=""):
    """Remove prefix and space, or only prefix, when possible"""
    if prefix:
        prefix_and_space = prefix + " "
        length_prefix = len(prefix)
        length_prefix_and_space = len(prefix_and_space)
        lines = [
            line[length_prefix_and_space:]
            if line.startswith(prefix_and_space)
            else (line[length_prefix:] if line.startswith(prefix) else line)
            for line in lines
        ]

    if suffix:
        space_and_suffix = " " + suffix
        length_suffix = len(suffix)
        length_space_and_suffix = len(space_and_suffix)
        lines = [
            line[:-length_space_and_suffix]
            if line.endswith(space_and_suffix)
            else (line[:-length_suffix] if line.endswith(suffix) else line)
            for line in lines
        ]

    return lines


def paragraph_is_fully_commented(lines, comment, main_language):
    """Is the paragraph fully commented?"""
    for i, line in enumerate(lines):
        if line.startswith(comment):
            if line[len(comment) :].lstrip().startswith(comment):
                continue
            if is_magic(line, main_language):
                return False
            continue
        return i > 0 and _BLANK_LINE.match(line)
    return True


def next_code_is_indented(lines):
    """Is the next unescaped line indented?"""
    for line in lines:
        if _BLANK_LINE.match(line):
            continue
        return _PY_INDENTED.match(line)
    return False


def count_lines_to_next_cell(cell_end_marker, next_cell_start, total, explicit_eoc):
    """How many blank lines between end of cell marker and next cell?"""
    if cell_end_marker < total:
        lines_to_next_cell = next_cell_start - cell_end_marker
        if explicit_eoc:
            lines_to_next_cell -= 1
        if next_cell_start >= total:
            lines_to_next_cell += 1
        return lines_to_next_cell

    return 1


def last_two_lines_blank(source):
    """Are the two last lines blank, and not the third last one?"""
    if len(source) < 3:
        return False
    return (
        not _BLANK_LINE.match(source[-3])
        and _BLANK_LINE.match(source[-2])
        and _BLANK_LINE.match(source[-1])
    )


class BaseCellReader:
    """A class that can read notebook cells from their text representation"""

    default_comment_magics = None
    lines_to_next_cell = 1

    start_code_re = None
    simple_start_code_re = None
    end_code_re = None

    # How to make code inactive
    comment = ""

    # Any specific prefix for lines in markdown cells (like in R spin format?)
    markdown_prefix = None

    def __init__(self, fmt=None, default_language=None):
        """Create a cell reader with empty content"""
        if not fmt:
            fmt = {}
        self.ext = fmt.get("extension")
        self.default_language = default_language or _SCRIPT_EXTENSIONS.get(
            self.ext, {}
        ).get("language", "python")
        self.comment_magics = fmt.get("comment_magics", self.default_comment_magics)
        self.use_runtools = fmt.get("use_runtools", False)
        self.format_version = fmt.get("format_version")
        self.metadata = None
        self.org_content = []
        self.content = []
        self.explicit_soc = None
        self.explicit_eoc = None
        self.cell_type = None
        self.language = None
        self.cell_metadata_json = fmt.get("cell_metadata_json", False)
        self.doxygen_equation_markers = fmt.get("doxygen_equation_markers", False)

    def read(self, lines):
        """Read one cell from the given lines, and return the cell,
        plus the position of the next cell
        """

        # Do we have an explicit code marker on the first line?
        self.metadata_and_language_from_option_line(lines[0])

        if self.metadata and "language" in self.metadata:
            self.language = self.metadata.pop("language")

        # Parse cell till its end and set content, lines_to_next_cell
        pos_next_cell = self.find_cell_content(lines)

        if self.cell_type == "code":
            new_cell = new_code_cell
        elif self.cell_type == "markdown":
            new_cell = new_markdown_cell
        else:
            new_cell = new_raw_cell

        if not self.metadata:
            self.metadata = {}

        if self.ext == ".py":
            expected_blank_lines = pep8_lines_between_cells(
                self.org_content or [""], lines[pos_next_cell:], self.ext
            )
        else:
            expected_blank_lines = 1

        if self.lines_to_next_cell != expected_blank_lines:
            self.metadata["lines_to_next_cell"] = self.lines_to_next_cell

        if self.language:
            self.metadata["language"] = self.language

        return (
            new_cell(source="\n".join(self.content), metadata=self.metadata),
            pos_next_cell,
        )

    def metadata_and_language_from_option_line(self, line):
        """Parse code options on the given line. When a start of a code cell
        is found, self.metadata is set to a dictionary."""
        if self.start_code_re.match(line):
            self.language, self.metadata = self.options_to_metadata(
                self.start_code_re.findall(line)[0]
            )

    def options_to_metadata(self, options):
        """Return language (str) and metadata (dict) from the option line"""
        raise NotImplementedError("Option parsing must be implemented in a sub class")

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        raise NotImplementedError("This method must be implemented in a sub class")

    def find_cell_content(self, lines):
        """Parse cell till its end and set content, lines_to_next_cell.
        Return the position of next cell start"""
        cell_end_marker, next_cell_start, self.explicit_eoc = self.find_cell_end(lines)

        # Metadata to dict
        if self.metadata is None:
            cell_start = 0
            self.metadata = {}
        else:
            cell_start = 1

        # Cell content
        source = lines[cell_start:cell_end_marker]
        self.org_content = copy(source)

        # Exactly two empty lines at the end of cell (caused by PEP8)?
        if self.ext == ".py" and self.explicit_eoc:
            if last_two_lines_blank(source):
                source = source[:-2]
                lines_to_end_of_cell_marker = 2
            else:
                lines_to_end_of_cell_marker = 0

            pep8_lines = pep8_lines_between_cells(
                source, lines[cell_end_marker:], self.ext
            )
            if lines_to_end_of_cell_marker != (0 if pep8_lines == 1 else 2):
                self.metadata[
                    "lines_to_end_of_cell_marker"
                ] = lines_to_end_of_cell_marker

        # Uncomment content
        self.explicit_soc = cell_start > 0
        self.content = self.extract_content(source)

        # Is this an inactive cell?
        if self.cell_type == "code":
            if not is_active(".ipynb", self.metadata):
                if self.metadata.get("active") == "":
                    del self.metadata["active"]
                self.cell_type = "raw"
            elif self.ext in [".md", ".markdown"] and not self.language:
                # Markdown files in version >= 1.3 represent code chunks with no language as Markdown cells
                if self.format_version not in ["1.0", "1.1"]:
                    self.cell_type = "markdown"
                    self.explicit_eoc = False
                    cell_end_marker += 1
                    self.content = lines[:cell_end_marker]
                # Previous versions mapped those to raw cells
                else:
                    self.cell_type = "raw"

        # Explicit end of cell marker?
        if (
            next_cell_start + 1 < len(lines)
            and _BLANK_LINE.match(lines[next_cell_start])
            and not _BLANK_LINE.match(lines[next_cell_start + 1])
        ):
            next_cell_start += 1
        elif (
            self.explicit_eoc
            and next_cell_start + 2 < len(lines)
            and _BLANK_LINE.match(lines[next_cell_start])
            and _BLANK_LINE.match(lines[next_cell_start + 1])
            and not _BLANK_LINE.match(lines[next_cell_start + 2])
        ):
            next_cell_start += 2

        self.lines_to_next_cell = count_lines_to_next_cell(
            cell_end_marker, next_cell_start, len(lines), self.explicit_eoc
        )

        return next_cell_start

    def uncomment_code_and_magics(self, lines):
        """Uncomment code and possibly commented magic commands"""
        raise NotImplementedError("This method must be implemented in a sub class")

    def extract_content(self, lines):
        # Code cells with just a multiline string become Markdown cells
        if self.ext == ".py" and not is_active(
            self.ext, self.metadata, self.cell_type == "code"
        ):
            content = "\n".join(lines).strip()
            for prefix in [""] if self.ext != ".py" else ["", "r", "R"]:
                for triple_quote in ['"""', "'''"]:
                    left = prefix + triple_quote
                    right = triple_quote
                    if (
                        content.startswith(left)
                        and content.endswith(right)
                        and len(content) >= len(left + right)
                    ):
                        content = content[len(left) : -len(right)]
                        # Trim first/last line return
                        if content.startswith("\n"):
                            content = content[1:]
                            left = left + "\n"
                        if content.endswith("\n"):
                            content = content[:-1]
                            right = "\n" + right

                        if not prefix:
                            if len(left) == len(right) == 4:
                                self.metadata["cell_marker"] = left[:3]
                        elif len(left[1:]) == len(right) == 4:
                            self.metadata["cell_marker"] = left[:4]
                        else:
                            self.metadata["cell_marker"] = left + "," + right
                        return content.splitlines()

        if not is_active(self.ext, self.metadata) or (
            "active" not in self.metadata
            and self.language
            and self.language != self.default_language
        ):
            return uncomment(
                lines, self.comment if self.ext not in [".r", ".R"] else "#"
            )

        return self.uncomment_code_and_magics(lines)


class MarkdownCellReader(BaseCellReader):
    """Read notebook cells from Markdown documents"""

    comment = ""
    start_code_re = re.compile(
        r"^```(`*)(\s*)({})($|\s.*$)".format(
            "|".join(_JUPYTER_LANGUAGES_LOWER_AND_UPPER).replace("+", "\\+")
        )
    )
    non_jupyter_code_re = re.compile(r"^```")
    end_code_re = re.compile(r"^```\s*$")
    start_region_re = re.compile(r"^<!--\s*#(region|markdown|md|raw)(.*)-->\s*$")
    end_region_re = None
    default_comment_magics = False

    def __init__(self, fmt=None, default_language=None):
        super().__init__(fmt, default_language)
        self.split_at_heading = (fmt or {}).get("split_at_heading", False)
        self.in_region = False
        self.in_raw = False
        if self.format_version in ["1.0", "1.1"] and self.ext != ".Rmd":
            # Restore the pattern used in Markdown <= 1.1
            self.start_code_re = re.compile(r"^```(.*)")
            self.non_jupyter_code_re = re.compile(r"^```\{")

    def metadata_and_language_from_option_line(self, line):
        match_region = self.start_region_re.match(line)
        if match_region:
            self.in_region = True
            groups = match_region.groups()
            region_name = groups[0]
            self.end_region_re = re.compile(rf"^<!--\s*#end{region_name}\s*-->\s*$")
            self.cell_metadata_json = self.cell_metadata_json or is_json_metadata(
                groups[1]
            )
            title, self.metadata = text_to_metadata(groups[1], allow_title=True)
            if region_name == "raw":
                self.cell_type = "raw"
            else:
                self.cell_type = "markdown"
            if title:
                self.metadata["title"] = title
            if region_name in ["markdown", "md"]:
                self.metadata["region_name"] = region_name
        elif self.start_code_re.match(line):
            self.language, self.metadata = self.options_to_metadata(
                self.start_code_re.findall(line)[0]
            )
            # Cells with a .noeval attribute are markdown cells #347
            if self.metadata.get(".noeval", "") is None:
                self.cell_type = "markdown"
                self.metadata = {}
                self.language = None

    def options_to_metadata(self, options):
        if isinstance(options, tuple):
            self.end_code_re = re.compile("```" + options[0])
            options = " ".join(options[1:])
        else:
            self.end_code_re = re.compile(r"^```\s*$")
        self.cell_metadata_json = self.cell_metadata_json or is_json_metadata(options)
        return text_to_metadata(options)

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        if self.in_region:
            for i, line in enumerate(lines):
                if self.end_region_re.match(line):
                    return i, i + 1, True
        elif self.metadata is None:
            # default markdown: (last) two consecutive blank lines, except when in code blocks
            self.cell_type = "markdown"
            prev_blank = 0
            in_explicit_code_block = False
            in_indented_code_block = False

            for i, line in enumerate(lines):
                if in_explicit_code_block and self.end_code_re.match(line):
                    in_explicit_code_block = False
                    continue

                if (
                    prev_blank
                    and line.startswith("    ")
                    and not _BLANK_LINE.match(line)
                ):
                    in_indented_code_block = True
                    prev_blank = 0
                    continue

                if (
                    in_indented_code_block
                    and not _BLANK_LINE.match(line)
                    and not line.startswith("    ")
                ):
                    in_indented_code_block = False

                if in_indented_code_block or in_explicit_code_block:
                    continue

                if self.start_region_re.match(line):
                    if i > 1 and prev_blank:
                        return i - 1, i, False
                    return i, i, False

                if self.start_code_re.match(line):
                    if line.startswith("```{bibliography}"):
                        in_explicit_code_block = True
                        prev_blank = 0
                        continue

                    # Cells with a .noeval attribute are markdown cells #347
                    _, metadata = self.options_to_metadata(
                        self.start_code_re.findall(line)[0]
                    )
                    if metadata.get(".noeval", "") is None:
                        in_explicit_code_block = True
                        prev_blank = 0
                        continue

                    if i > 1 and prev_blank:
                        return i - 1, i, False
                    return i, i, False

                if self.non_jupyter_code_re.match(line):
                    if prev_blank >= 2:
                        return i - 2, i, True
                    in_explicit_code_block = True
                    prev_blank = 0
                    continue

                if self.split_at_heading and line.startswith("#") and prev_blank >= 1:
                    return i - 1, i, False

                if _BLANK_LINE.match(lines[i]):
                    prev_blank += 1
                elif prev_blank >= 2:
                    return i - 2, i, True
                else:
                    prev_blank = 0
        else:
            self.cell_type = "code"
            # At some point we could remove the below, in which we make sure not to break language strings
            # into multiple cells (#419). Indeed, now that the markdown cell uses one extra backtick (#712)
            # we should not have the issue any more
            parser = StringParser(self.language or self.default_language)
            for i, line in enumerate(lines):
                # skip cell header
                if i == 0:
                    continue

                if parser.is_quoted():
                    parser.read_line(line)
                    continue

                parser.read_line(line)
                if self.end_code_re.match(line):
                    return i, i + 1, True

        # End not found
        return len(lines), len(lines), False

    def uncomment_code_and_magics(self, lines):
        if self.cell_type == "code" and self.comment_magics:
            lines = uncomment_magic(lines, self.language)
        if self.cell_type == "markdown" and self.doxygen_equation_markers:
            lines = doxygen_to_markdown("\n".join(lines)).splitlines()
        return lines


class RMarkdownCellReader(MarkdownCellReader):
    """Read notebook cells from R Markdown notebooks"""

    comment = ""
    start_code_re = re.compile(r"^```{(.*)}\s*$")
    non_jupyter_code_re = re.compile(r"^```([^\{]|\s*$)")
    default_language = "R"
    default_comment_magics = True

    def options_to_metadata(self, options):
        return rmd_options_to_metadata(options, self.use_runtools)

    def uncomment_code_and_magics(self, lines):
        if (
            self.cell_type == "code"
            and self.comment_magics
            and is_active(self.ext, self.metadata)
        ):
            uncomment_magic(lines, self.language or self.default_language)

        return lines


class ScriptCellReader(BaseCellReader):  # pylint: disable=W0223
    """Read notebook cells from scripts
    (common base for R and Python scripts)"""

    def uncomment_code_and_magics(self, lines):
        if self.cell_type == "code" or self.comment != "#'":
            if self.comment_magics:
                if is_active(self.ext, self.metadata):
                    uncomment_magic(
                        lines,
                        self.language or self.default_language,
                        explicitly_code=self.explicit_soc,
                    )
                    if (
                        self.cell_type == "code"
                        and not self.explicit_soc
                        and need_explicit_marker(
                            lines, self.language or self.default_language
                        )
                    ):
                        self.metadata["comment_questions"] = False
                else:
                    lines = uncomment(lines)

        if self.default_language == "go" and self.language is None:
            lines = [
                re.sub(r"^((//\s*)*)(//\s*gonb:%%)", r"\1%%", line) for line in lines
            ]

        if self.cell_type == "code":
            return unescape_code_start(
                lines, self.ext, self.language or self.default_language
            )

        return uncomment(
            lines, self.markdown_prefix or self.comment, self.comment_suffix
        )


class RScriptCellReader(ScriptCellReader):
    """Read notebook cells from R scripts written according
    to the knitr-spin syntax"""

    comment = "#'"
    comment_suffix = ""
    markdown_prefix = "#'"
    default_language = "R"
    start_code_re = re.compile(r"^#\+(.*)\s*$")
    default_comment_magics = True

    def options_to_metadata(self, options):
        return rmd_options_to_metadata("r " + options, self.use_runtools)

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""
        if self.metadata is None and lines[0].startswith("#'"):
            self.cell_type = "markdown"
            for i, line in enumerate(lines):
                if not line.startswith("#'"):
                    if _BLANK_LINE.match(line):
                        return i, i + 1, False
                    return i, i, False

            return len(lines), len(lines), False

        if self.metadata and "cell_type" in self.metadata:
            self.cell_type = self.metadata.pop("cell_type")
        else:
            self.cell_type = "code"

        parser = StringParser(self.language or self.default_language)
        for i, line in enumerate(lines):
            # skip cell header
            if self.metadata is not None and i == 0:
                continue

            if parser.is_quoted():
                parser.read_line(line)
                continue

            parser.read_line(line)

            if self.start_code_re.match(line) or (
                self.markdown_prefix and line.startswith(self.markdown_prefix)
            ):
                if i > 0 and _BLANK_LINE.match(lines[i - 1]):
                    if i > 1 and _BLANK_LINE.match(lines[i - 2]):
                        return i - 2, i, False
                    return i - 1, i, False
                return i, i, False

            if _BLANK_LINE.match(line):
                if not next_code_is_indented(lines[i:]):
                    if i > 0:
                        return i, i + 1, False
                    if len(lines) > 1 and not _BLANK_LINE.match(lines[1]):
                        return 1, 1, False
                    return 1, 2, False

        return len(lines), len(lines), False


class LightScriptCellReader(ScriptCellReader):
    """Read notebook cells from plain Python or Julia files. Cells
    are identified by line breaks, unless they start with an
    explicit marker '# +'"""

    default_comment_magics = True
    cell_marker_start = None
    cell_marker_end = None

    def __init__(self, fmt=None, default_language=None):
        super().__init__(fmt, default_language)
        self.ext = self.ext or ".py"
        script = _SCRIPT_EXTENSIONS[self.ext]
        self.default_language = default_language or script["language"]
        self.comment = script["comment"]
        self.comment_suffix = script.get("comment_suffix", "")
        self.ignore_end_marker = True
        self.explicit_end_marker_required = False
        if (
            fmt
            and fmt.get("format_name", "light") == "light"
            and "cell_markers" in fmt
            and fmt["cell_markers"] != "+,-"
        ):
            self.cell_marker_start, self.cell_marker_end = fmt["cell_markers"].split(
                ",", 1
            )
            self.start_code_re = re.compile(
                "^"
                + re.escape(self.comment)
                + r"\s*"
                + self.cell_marker_start
                + r"(.*)$"
            )
            self.end_code_re = re.compile(
                "^" + re.escape(self.comment) + r"\s*" + self.cell_marker_end + r"\s*$"
            )
        else:
            self.start_code_re = re.compile(
                "^" + re.escape(self.comment) + r"\s*\+(.*)$"
            )

    def metadata_and_language_from_option_line(self, line):
        if self.start_code_re.match(line):
            # Remove the OCAML suffix
            if self.comment_suffix:
                if line.endswith(" " + self.comment_suffix):
                    line = line[: -len(" " + self.comment_suffix)]
                elif line.endswith(self.comment_suffix):
                    line = line[: -len(self.comment_suffix)]

            # We want to parse inner most regions as cells.
            # Thus, if we find another region start before the end for this region,
            # we will have ignore the metadata that we found here, and move on to the next cell.
            groups = self.start_code_re.match(line).groups()
            self.language, self.metadata = self.options_to_metadata(groups[0])
            self.ignore_end_marker = False
            if self.cell_marker_start:
                self.explicit_end_marker_required = True
        elif self.simple_start_code_re and self.simple_start_code_re.match(line):
            self.metadata = {}
            self.ignore_end_marker = False
        elif self.cell_marker_end and self.end_code_re.match(line):
            self.metadata = None
            self.cell_type = "code"

    def options_to_metadata(self, options):
        self.cell_metadata_json = self.cell_metadata_json or is_json_metadata(options)
        title, metadata = text_to_metadata(options, allow_title=True)

        # Cell type
        for cell_type in ["markdown", "raw", "md"]:
            code = f"[{cell_type}]"
            if code in title:
                title = title.replace(code, "").strip()
                metadata["cell_type"] = cell_type
                if cell_type == "md":
                    metadata["region_name"] = cell_type
                    metadata["cell_type"] = "markdown"
                break

        # Spyder has sub cells
        cell_depth = 0
        while title.startswith("%"):
            cell_depth += 1
            title = title[1:]

        if cell_depth:
            metadata["cell_depth"] = cell_depth
            title = title.strip()

        if title:
            metadata["title"] = title

        return None, metadata

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position of first line after cell"""
        if (
            self.metadata is None
            and not (self.cell_marker_end and self.end_code_re.match(lines[0]))
            and paragraph_is_fully_commented(lines, self.comment, self.default_language)
        ):
            self.cell_type = "markdown"
            for i, line in enumerate(lines):
                if _BLANK_LINE.match(line):
                    return i, i + 1, False
            return len(lines), len(lines), False

        if self.metadata is None:
            self.end_code_re = None
        elif not self.cell_marker_end:
            end_of_cell = self.metadata.get("endofcell", "-")
            self.end_code_re = re.compile(
                "^" + re.escape(self.comment) + " " + end_of_cell + r"\s*$"
            )

        return self.find_region_end(lines)

    def find_region_end(self, lines):
        """Find the end of the region started with start and end markers"""
        if self.metadata and "cell_type" in self.metadata:
            self.cell_type = self.metadata.pop("cell_type")
        else:
            self.cell_type = "code"

        parser = StringParser(self.language or self.default_language)
        for i, line in enumerate(lines):
            # skip cell header
            if self.metadata is not None and i == 0:
                continue

            if parser.is_quoted():
                parser.read_line(line)
                continue

            parser.read_line(line)
            # New code region
            # Simple code pattern in LightScripts must be preceded with a blank line
            if self.start_code_re.match(line) or (
                self.simple_start_code_re
                and self.simple_start_code_re.match(line)
                and (
                    self.cell_marker_start or i == 0 or _BLANK_LINE.match(lines[i - 1])
                )
            ):
                if self.explicit_end_marker_required:
                    # Metadata here was conditioned on finding an explicit end marker
                    # before the next start marker. So we dismiss it.
                    self.metadata = None
                    self.language = None

                if i > 0 and _BLANK_LINE.match(lines[i - 1]):
                    if i > 1 and _BLANK_LINE.match(lines[i - 2]):
                        return i - 2, i, False
                    return i - 1, i, False
                return i, i, False

            if not self.ignore_end_marker and self.end_code_re:
                if self.end_code_re.match(line):
                    return i, i + 1, True
            elif _BLANK_LINE.match(line):
                if not next_code_is_indented(lines[i:]):
                    if i > 0:
                        return i, i + 1, False
                    if len(lines) > 1 and not _BLANK_LINE.match(lines[1]):
                        return 1, 1, False
                    return 1, 2, False

        return len(lines), len(lines), False


class DoublePercentScriptCellReader(LightScriptCellReader):
    """Read notebook cells from Spyder/VScode scripts (#59)"""

    default_comment_magics = True

    def __init__(self, fmt, default_language=None):
        LightScriptCellReader.__init__(self, fmt, default_language)
        script = _SCRIPT_EXTENSIONS[self.ext]
        self.default_language = default_language or script["language"]
        self.comment = script["comment"]
        self.comment_suffix = script.get("comment_suffix", "")
        self.start_code_re = re.compile(
            rf"^\s*{re.escape(self.comment)}\s*%%(%*)\s(.*)$"
        )
        self.alternative_start_code_re = re.compile(
            r"^\s*{}\s*(%%|<codecell>|In\[[0-9 ]*\]:?)\s*$".format(
                re.escape(self.comment)
            )
        )
        self.explicit_soc = True

    def metadata_and_language_from_option_line(self, line):
        """Parse code options on the given line. When a start of a code cell
        is found, self.metadata is set to a dictionary."""
        if self.start_code_re.match(line):
            line = uncomment([line], self.comment, self.comment_suffix)[0]
            self.language, self.metadata = self.options_to_metadata(
                line[line.find("%%") + 2 :]
            )
        else:
            self.metadata = {}

    def find_cell_content(self, lines):
        """Parse cell till its end and set content, lines_to_next_cell.
        Return the position of next cell start"""
        cell_end_marker, next_cell_start, explicit_eoc = self.find_cell_end(lines)

        # Metadata to dict
        if self.start_code_re.match(lines[0]) or self.alternative_start_code_re.match(
            lines[0]
        ):
            cell_start = 1
        else:
            cell_start = 0

        # Cell content
        source = lines[cell_start:cell_end_marker]
        self.org_content = copy(source)
        self.content = self.extract_content(source)

        self.lines_to_next_cell = count_lines_to_next_cell(
            cell_end_marker, next_cell_start, len(lines), explicit_eoc
        )

        return next_cell_start

    def find_cell_end(self, lines):
        """Return position of end of cell marker, and position
        of first line after cell"""

        if self.metadata and "cell_type" in self.metadata:
            self.cell_type = self.metadata.pop("cell_type")
        elif not is_active(".ipynb", self.metadata):
            if self.metadata.get("active") == "":
                del self.metadata["active"]
            self.cell_type = "raw"
            if is_active(self.ext, self.metadata):
                self.comment = ""
        else:
            self.cell_type = "code"

        next_cell = len(lines)
        parser = StringParser(self.language or self.default_language)
        for i, line in enumerate(lines):
            if parser.is_quoted():
                parser.read_line(line)
                continue

            parser.read_line(line)
            if i > 0 and (
                self.start_code_re.match(line)
                or self.alternative_start_code_re.match(line)
            ):
                next_cell = i
                break

        if last_two_lines_blank(lines[:next_cell]):
            return next_cell - 2, next_cell, False
        if next_cell > 0 and _BLANK_LINE.match(lines[next_cell - 1]):
            return next_cell - 1, next_cell, False
        return next_cell, next_cell, False


class HydrogenCellReader(DoublePercentScriptCellReader):
    """Read notebook cells from Hydrogen scripts (#59)"""

    default_comment_magics = False


class SphinxGalleryScriptCellReader(ScriptCellReader):  # pylint: disable=W0223
    """Read notebook cells from Sphinx Gallery scripts (#80)"""

    comment = "#"
    default_language = "python"
    default_comment_magics = True
    twenty_hash = re.compile(r"^#( |)#{19,}\s*$")
    default_markdown_cell_marker = "#" * 79
    markdown_marker = None

    def __init__(self, fmt=None, default_language="python"):
        super().__init__(fmt, default_language)
        self.ext = ".py"
        self.rst2md = (fmt or {}).get("rst2md", False)

    def start_of_new_markdown_cell(self, line):
        """Does this line starts a new markdown cell?
        Then, return the cell marker"""
        for empty_markdown_cell in ['""', "''"]:
            if line == empty_markdown_cell:
                return empty_markdown_cell

        for triple_quote in ['"""', "'''"]:
            if line.startswith(triple_quote):
                return triple_quote

        if self.twenty_hash.match(line):
            return line

        return None

    def metadata_and_language_from_option_line(self, line):
        self.markdown_marker = self.start_of_new_markdown_cell(line)
        if self.markdown_marker:
            self.cell_type = "markdown"
            if self.markdown_marker != self.default_markdown_cell_marker:
                self.metadata = {"cell_marker": self.markdown_marker}
        else:
            self.cell_type = "code"

    def find_cell_end(self, lines):
        """Return position of end of cell, and position
        of first line after cell, and whether there was an
        explicit end of cell marker"""

        if self.cell_type == "markdown":
            # Empty cell "" or ''
            if len(self.markdown_marker) <= 2:
                if len(lines) == 1 or _BLANK_LINE.match(lines[1]):
                    return 0, 2, True
                return 0, 1, True

            # Multi-line comment with triple quote
            if len(self.markdown_marker) == 3:
                for i, line in enumerate(lines):
                    if (
                        i > 0 or line.strip() != self.markdown_marker
                    ) and line.rstrip().endswith(self.markdown_marker):
                        explicit_end_of_cell_marker = (
                            line.strip() == self.markdown_marker
                        )
                        if explicit_end_of_cell_marker:
                            end_of_cell = i
                        else:
                            end_of_cell = i + 1
                        if len(lines) <= i + 1 or _BLANK_LINE.match(lines[i + 1]):
                            return end_of_cell, i + 2, explicit_end_of_cell_marker
                        return end_of_cell, i + 1, explicit_end_of_cell_marker
            else:
                # 20 # or more
                for i, line in enumerate(lines[1:], 1):
                    if not line.startswith(self.comment):
                        if _BLANK_LINE.match(line):
                            return i, i + 1, False
                        return i, i, False

        elif self.cell_type == "code":
            parser = StringParser("python")
            for i, line in enumerate(lines):
                if parser.is_quoted():
                    parser.read_line(line)
                    continue

                if self.start_of_new_markdown_cell(line):
                    if i > 0 and _BLANK_LINE.match(lines[i - 1]):
                        return i - 1, i, False
                    return i, i, False
                parser.read_line(line)

        return len(lines), len(lines), False

    def find_cell_content(self, lines):
        """Parse cell till its end and set content, lines_to_next_cell.
        Return the position of next cell start"""
        cell_end_marker, next_cell_start, explicit_eoc = self.find_cell_end(lines)

        # Metadata to dict
        cell_start = 0
        if self.cell_type == "markdown":
            if self.markdown_marker in ['"""', "'''"]:
                # Remove the triple quotes
                if lines[0].strip() == self.markdown_marker:
                    cell_start = 1
                else:
                    lines[0] = lines[0][3:]
                if not explicit_eoc:
                    last = lines[cell_end_marker - 1]
                    lines[cell_end_marker - 1] = last[
                        : last.rfind(self.markdown_marker)
                    ]
            if self.twenty_hash.match(self.markdown_marker):
                cell_start = 1
        else:
            self.metadata = {}

        # Cell content
        source = lines[cell_start:cell_end_marker]
        self.org_content = copy(source)

        if self.cell_type == "code" and self.comment_magics:
            uncomment_magic(source, self.language or self.default_language)

        if self.cell_type == "markdown" and source:
            if self.markdown_marker.startswith(self.comment):
                source = uncomment(source, self.comment)
            if self.rst2md:
                if rst2md:
                    gallery_conf = {"notebook_images": False}
                    heading_level_counter = count(start=1)
                    heading_levels = defaultdict(lambda: next(heading_level_counter))
                    source = rst2md(
                        "\n".join(source), gallery_conf, "", heading_levels
                    ).splitlines()
                else:
                    raise ImportError(
                        "Could not import rst2md from sphinx_gallery.notebook"
                    )  # pragma: no cover

        self.content = source

        self.lines_to_next_cell = count_lines_to_next_cell(
            cell_end_marker, next_cell_start, len(lines), explicit_eoc
        )

        return next_cell_start



""" Main for Jupytext_config

Call with (e.g.)::

    python -m jupytext_config list-default-viewer
"""

import sys

from .jupytext_config import main

if __name__ == "__main__":
    sys.exit(main())



"""Jupytext should refuse to open a file with invalid content"""

from pathlib import Path

import pytest
from jupyter_server.utils import ensure_async
from tornado.web import HTTPError

from jupytext import read
from jupytext.cli import jupytext as jupytext_cli


@pytest.fixture
def invalid_md_file():
    return Path(__file__).parent / "invalid_file_896.md"


@pytest.mark.skip_on_windows
def test_read_invalid_md_file_fails(invalid_md_file):
    with open(invalid_md_file) as fp:
        with pytest.raises(UnicodeDecodeError):
            read(fp)


def test_convert_invalid_md_file_fails(invalid_md_file):
    with pytest.raises(UnicodeDecodeError):
        jupytext_cli(["--to", "ipynb", str(invalid_md_file)])


@pytest.mark.asyncio
async def test_open_invalid_md_file_fails(invalid_md_file, tmp_path, cm):
    cm.root_dir = str(invalid_md_file.parent)

    with pytest.raises(HTTPError, match="invalid_file_896.md is not UTF-8 encoded"):
        await ensure_async(cm.get(invalid_md_file.name))



import logging
import os
import sys
import unittest.mock as mock

import pytest
from jupyter_server.utils import ensure_async
from nbformat import read
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook
from tornado.web import HTTPError

import jupytext
from jupytext.compare import compare_cells, notebook_model

SAMPLE_NOTEBOOK = new_notebook(
    cells=[new_markdown_cell("A Markdown cell"), new_code_cell("# A code cell\n1 + 1")]
)

pytestmark = pytest.mark.asyncio


async def test_local_config_overrides_cm_config(tmpdir, cm):
    cm.root_dir = str(tmpdir)
    cm.formats = "ipynb,py"

    nested = tmpdir.mkdir("nested")
    with open(str(nested.join("jupytext.yml")), "w") as fp:
        fp.write("formats: ''\n")

    await ensure_async(cm.save(notebook_model(SAMPLE_NOTEBOOK), "notebook.ipynb"))
    assert os.path.isfile(str(tmpdir.join("notebook.ipynb")))
    assert os.path.isfile(str(tmpdir.join("notebook.py")))

    await ensure_async(
        cm.save(notebook_model(SAMPLE_NOTEBOOK), "nested/notebook.ipynb")
    )
    assert os.path.isfile(str(nested.join("notebook.ipynb")))
    assert not os.path.isfile(str(nested.join("notebook.py")))


async def test_config_file_is_called_just_once(tmpdir, cm, n=2):
    cm.root_dir = str(tmpdir)
    tmpdir.join("jupytext.toml").write("")
    nb_files = [str(tmpdir.join(f"notebook{i}.ipynb")) for i in range(n)]

    for nb_file in nb_files:
        jupytext.write(SAMPLE_NOTEBOOK, nb_file)

    mock_config = mock.MagicMock(return_value=None)

    with mock.patch(
        "jupytext.sync_contentsmanager.load_jupytext_configuration_file", mock_config
    ):
        with mock.patch(
            "jupytext.async_contentsmanager.load_jupytext_configuration_file",
            mock_config,
        ):
            for i in range(n):
                await ensure_async(cm.get(f"notebook{i}.ipynb", content=False))

    # Listing the contents should not call the config more than once
    assert mock_config.call_count == 1


async def test_pairing_through_config_leaves_ipynb_unmodified(tmpdir, cm):
    cm.root_dir = str(tmpdir)

    cfg_file = tmpdir.join("jupytext.yml")
    nb_file = tmpdir.join("notebook.ipynb")
    py_file = tmpdir.join("notebook.py")

    cfg_file.write("formats: 'ipynb,py'\n")

    await ensure_async(cm.save(notebook_model(SAMPLE_NOTEBOOK), "notebook.ipynb"))
    assert nb_file.isfile()
    assert py_file.isfile()

    nb = read(nb_file, as_version=4)
    assert "jupytext" not in nb.metadata


@pytest.mark.parametrize(
    "cfg_file,cfg_text",
    [
        # Should be false, not False
        ("jupytext.toml", "hide_notebook_metadata = False"),
        ("jupytext.toml", 'hide_notebook_metadata = "False"'),
        ("jupytext.toml", "not_a_jupytext_option = true"),
        ("pyproject.toml", "[tool.jupytext]\nnot_a_jupytext_option = true"),
        ("jupytext.json", '{"notebook_metadata_filter":"-all",}'),
    ],
)
@pytest.mark.filterwarnings(
    r"ignore:Passing (unrecognized|unrecoginized) arguments "
    r"to super\(JupytextConfiguration\).__init__"
)
async def test_incorrect_config_message(tmpdir, cfg_file, cfg_text, cm):
    cm.root_dir = str(tmpdir)

    tmpdir.join(cfg_file).write(cfg_text)
    tmpdir.join("empty.ipynb").write("{}")

    expected_message = "The Jupytext configuration file .*{} is incorrect".format(
        cfg_file
    )

    with pytest.raises(HTTPError, match=expected_message):
        await ensure_async(cm.get("empty.ipynb", type="notebook", content=False))

    with pytest.raises(HTTPError, match=expected_message):
        await ensure_async(cm.save(notebook_model(SAMPLE_NOTEBOOK), "notebook.ipynb"))


async def test_global_config_file(tmpdir, cm):
    cm_dir = tmpdir.join("cm_dir").mkdir()
    cm.root_dir = str(cm_dir)

    tmpdir.join("jupytext.toml").write('formats = "ipynb,Rmd"')

    def fake_global_config_directory():
        return [str(tmpdir)]

    with mock.patch(
        "jupytext.config.global_jupytext_configuration_directories",
        fake_global_config_directory,
    ):
        nb = new_notebook(cells=[new_code_cell("1+1")])
        model = notebook_model(nb)
        await ensure_async(cm.save(model, "notebook.ipynb"))
        assert {
            model["path"]
            for model in (await ensure_async(cm.get("/", content=True)))["content"]
        } == {
            "notebook.ipynb",
            "notebook.Rmd",
        }


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="AttributeError: 'LocalPath' object has no attribute 'mksymlinkto'",
)
async def test_paired_files_and_symbolic_links(tmpdir, cm):
    """We test that we don't get issues when pairing files into folders
    that are symbolic links"""

    actual = tmpdir.mkdir("actual_files")
    actual_notebooks = actual.mkdir("notebooks")
    actual_scripts = actual.mkdir("scripts")

    # Open the contents manager in another dir
    jupyter_dir = tmpdir.mkdir("jupyter_dir")
    cm.root_dir = str(jupyter_dir)

    # Create sym links to the notebook/script folders
    jupyter_dir.join("link_to_notebooks").mksymlinkto(actual_notebooks)
    jupyter_dir.join("link_to_scripts").mksymlinkto(actual_scripts)

    # Pair the notebooks in the linked folders
    jupyter_dir.join("jupytext.toml").write(
        'formats = "link_to_notebooks///ipynb,link_to_scripts///py:percent"'
    )

    # Save a notebook
    await ensure_async(
        cm.save(notebook_model(SAMPLE_NOTEBOOK), "link_to_notebooks/notebook.ipynb")
    )

    # This creates two files in the destinations folders
    assert actual_notebooks.join("notebook.ipynb").isfile()
    assert actual_scripts.join("notebook.py").isfile()

    # Re-open the notebook (here, the text version)
    await ensure_async(cm.get("link_to_scripts/notebook.py"))

    # Update the text version
    jupyter_dir.join("link_to_scripts").join("notebook.py").write_text(
        "# %%\n3 + 3\n", encoding="utf-8"
    )

    # Reload and make sure that we get the updated notebook
    model = await ensure_async(cm.get("link_to_notebooks/notebook.ipynb"))
    nb = model["content"]
    compare_cells(nb.cells, [new_code_cell("3 + 3")], compare_ids=False)


async def test_metadata_filter_from_config_has_precedence_over_notebook_metadata(
    tmpdir, cwd_tmpdir, cm, python_notebook
):
    python_notebook.metadata["jupytext"] = {"notebook_metadata_filter": "-all"}
    tmpdir.join("jupytext.toml").write('notebook_metadata_filter = "all"')

    cm.root_dir = str(tmpdir)

    await ensure_async(cm.save(notebook_model(python_notebook), "test.py"))

    py = tmpdir.join("test.py").read()
    assert "notebook_metadata_filter: all" in py


async def test_test_no_text_representation_metadata_in_ipynb_900(
    tmpdir, python_notebook, cm
):
    tmpdir.join("jupytext.toml").write('formats = "ipynb,py:percent"\n')

    # create a test notebook and save it in Jupyter
    nb = python_notebook
    cm.root_dir = str(tmpdir)
    await ensure_async(cm.save(dict(type="notebook", content=nb), "test.ipynb"))

    # Assert that "text_representation" is in the Jupytext metadata #900
    assert "text_representation" in tmpdir.join("test.py").read()
    # But not in the ipynb notebook
    assert "text_representation" not in tmpdir.join("test.ipynb").read()

    # modify the ipynb file in Jupyter
    # Reload the notebook
    nb = (await ensure_async(cm.get("test.ipynb")))["content"]
    nb.cells.append(new_markdown_cell("A new cell"))
    await ensure_async(cm.save(dict(type="notebook", content=nb), "test.ipynb"))

    # The text representation metadata is in the py file
    assert "text_representation" in tmpdir.join("test.py").read()
    # But not in the ipynb notebook
    assert "text_representation" not in tmpdir.join("test.ipynb").read()


async def test_cm_config_no_log(cwd_tmp_path, tmp_path, caplog, cm):
    cm.root_dir = str(tmp_path)

    config = 'cm_config_log_level="none"'
    (tmp_path / "jupytext.toml").write_text(config)
    (tmp_path / "nb1.py").write_text("# %%")
    (tmp_path / "subfolder").mkdir()
    (tmp_path / "subfolder" / "jupytext.toml").write_text(config)
    (tmp_path / "subfolder" / "nb2.py").write_text("# %%")

    caplog.set_level(logging.DEBUG)

    await ensure_async(cm.get("nb1.py", type="notebook", content=False))
    await ensure_async(cm.get("nb1.py", type="notebook", content=True))
    await ensure_async(cm.get("subfolder/nb2.py", type="notebook", content=False))
    await ensure_async(cm.get("subfolder/nb2.py", type="notebook", content=True))

    assert "Jupytext configuration file" not in caplog.text


async def test_cm_config_log_only_if_changed(cwd_tmp_path, tmp_path, caplog, cm):
    cm.root_dir = str(tmp_path)

    config = ""
    (tmp_path / "jupytext.toml").write_text(config)
    (tmp_path / "nb1.py").write_text("# %%")
    (tmp_path / "subfolder").mkdir()
    (tmp_path / "subfolder" / "jupytext.toml").write_text(config)
    (tmp_path / "subfolder" / "nb2.py").write_text("# %%")

    caplog.set_level(logging.INFO)

    await ensure_async(cm.get("nb1.py", type="notebook", content=False))
    assert "Jupytext configuration file" in caplog.text
    caplog.clear()

    # Same notebook, same config => no log
    await ensure_async(cm.get("nb1.py", type="notebook", content=True))
    assert "Jupytext configuration file" not in caplog.text

    # Same notebook, config changed => log
    (tmp_path / "jupytext.toml").write_text('formats="ipynb,py:percent"')
    await ensure_async(cm.get("nb1.py", type="notebook", content=True))
    assert "Jupytext configuration file" in caplog.text
    caplog.clear()

    # Different folder, different config
    await ensure_async(cm.get("subfolder/nb2.py", type="notebook", content=False))
    assert "Jupytext configuration file" in caplog.text
    caplog.clear()

    # Same config as previously => no log
    await ensure_async(cm.get("subfolder/nb2.py", type="notebook", content=False))
    assert "Jupytext configuration file" not in caplog.text



import jupytext


def test_incomplete_header(
    rmd="""---
title: Incomplete header

```{python}
1+1
```
""",
):
    nb = jupytext.reads(rmd, "Rmd")
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "---\ntitle: Incomplete header"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "1+1"


def test_code_in_markdown_block(
    rmd="""```{python}
a = 1
b = 2
a + b
```

```python
'''Code here goes to a Markdown cell'''


'''even if we have two blank lines above'''
```

```{bash}
ls -l
```
""",
):
    nb = jupytext.reads(rmd, "Rmd")
    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "code"
    assert nb.cells[0].source == "a = 1\nb = 2\na + b"
    assert nb.cells[1].cell_type == "markdown"
    assert (
        nb.cells[1].source
        == """```python
'''Code here goes to a Markdown cell'''


'''even if we have two blank lines above'''
```"""
    )
    assert nb.cells[2].cell_type == "code"
    assert nb.cells[2].source == "%%bash\nls -l"


def test_unterminated_header(
    rmd="""---
title: Unterminated header

```{python}
1+3
```

some text

```{r}
1+4
```

```{python not_terminated}
1+5
""",
):
    nb = jupytext.reads(rmd, "Rmd")
    assert len(nb.cells) == 5
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "---\ntitle: Unterminated header"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "1+3"
    assert nb.cells[2].cell_type == "markdown"
    assert nb.cells[2].source == "some text"
    assert nb.cells[3].cell_type == "code"
    assert nb.cells[3].source == "%%R\n1+4"
    assert nb.cells[4].cell_type == "code"
    assert nb.cells[4].metadata == {"name": "not_terminated"}
    assert nb.cells[4].source == "1+5"



# -*- coding: utf-8 -*-
"""
Notebook styled examples
========================

The gallery is capable of transforming Python files into reStructuredText files
with a notebook structure. For this to be used you need to respect some syntax
rules.

It makes a lot of sense to contrast this output rst file with the
:download:`original Python script <plot_notebook.py>` to get better feeling of
the necessary file structure.

Anything before the Python script docstring is ignored by sphinx-gallery and
will not appear in the rst file, nor will it be executed.
This Python docstring requires an reStructuredText title to name the file and
correctly build the reference links.

Once you close the docstring you would be writing Python code. This code gets
executed by sphinx gallery shows the plots and attaches the generating code.
Nevertheless you can break your code into blocks and give the rendered file
a notebook style. In this case you have to include a code comment breaker
a line of at least 20 hashes and then every comment start with the a new hash.

As in this example we start by first writing this module
style docstring, then for the first code block we write the example file author
and script license continued by the import modules instructions.

Original script from:
https://sphinx-gallery.readthedocs.io/en/latest/tutorials/plot_notebook.html
"""

# Code source: Óscar Nájera
# License: BSD 3 clause

import numpy as np
import matplotlib.pyplot as plt

##############################################################################
# This code block is executed, although it produces no output. Lines starting
# with a simple hash are code comment and get treated as part of the code
# block. To include this new comment string we started the new block with a
# long line of hashes.
#
# The sphinx-gallery parser will assume everything after this splitter and that
# continues to start with a **comment hash and space** (respecting code style)
# is text that has to be rendered in
# html format. Keep in mind to always keep your comments always together by
# comment hashes. That means to break a paragraph you still need to comment
# that line break.
#
# In this example the next block of code produces some plotable data. Code is
# executed, figure is saved and then code is presented next, followed by the
# inlined figure.

x = np.linspace(-np.pi, np.pi, 300)
xx, yy = np.meshgrid(x, x)
z = np.cos(xx) + np.cos(yy)

plt.figure()
plt.imshow(z)
plt.colorbar()
plt.xlabel('$x$')
plt.ylabel('$y$')

###########################################################################
# Again it is possible to continue the discussion with a new Python string. This
# time to introduce the next code block generates 2 separate figures.

plt.figure()
plt.imshow(z, cmap=plt.cm.get_cmap('hot'))
plt.figure()
plt.imshow(z, cmap=plt.cm.get_cmap('Spectral'), interpolation='none')

##########################################################################
# There's some subtle differences between rendered html rendered comment
# strings and code comment strings which I'll demonstrate below. (Some of this
# only makes sense if you look at the
# :download:`raw Python script <plot_notebook.py>`)
#
# Comments in comment blocks remain nested in the text.


def dummy():
    """Dummy function to make sure docstrings don't get rendered as text"""
    pass

# Code comments not preceded by the hash splitter are left in code blocks.

string = """
Triple-quoted string which tries to break parser but doesn't.
"""

############################################################################
# Output of the script is captured:

print('Some output from Python')

############################################################################
# Finally, I'll call ``show`` at the end just so someone running the Python
# code directly will see the plots; this is not necessary for creating the docs

plt.show()

############################################################################
# You can also include :math:`math` inline, or as separate equations:
#
# .. math::
#
#    \exp(j\pi) = -1
#
# You can also insert images:
#
# .. image:: http://www.sphinx-doc.org/en/stable/_static/sphinxheader.png
#    :alt: Sphinx header
#



import pytest
from git.exc import HookExecutionError
from pre_commit.main import main as pre_commit

from jupytext import read, write
from jupytext.cli import jupytext


def test_pre_commit_hook_sync_nbstripout(
    tmpdir,
    cwd_tmpdir,
    tmp_repo,
    jupytext_repo_root,
    jupytext_repo_rev,
    notebook_with_outputs,
):
    """Here we sync the ipynb notebook with a Markdown file and also apply nbstripout."""
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [--sync]

- repo: https://github.com/kynan/nbstripout
  rev: 0.5.0
  hooks:
  - id: nbstripout
"""

    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)

    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks", "-f"])

    # write a test notebook
    write(notebook_with_outputs, "test.ipynb")

    # We pair the notebook to a md file
    jupytext(["--set-formats", "ipynb,md", "test.ipynb"])

    # try to commit it, should fail because
    # 1. the md version hasn't been added
    # 2. the notebook has outputs
    tmp_repo.git.add("test.ipynb")
    with pytest.raises(
        HookExecutionError,
        match="files were modified by this hook",
    ):
        tmp_repo.index.commit("failing")

    # Add the two files
    tmp_repo.git.add("test.ipynb")
    tmp_repo.git.add("test.md")

    # now the commit will succeed
    tmp_repo.index.commit("passing")
    assert "test.ipynb" in tmp_repo.tree()
    assert "test.md" in tmp_repo.tree()

    # the ipynb file has no outputs on disk
    nb = read("test.ipynb")
    assert not nb.cells[0].outputs



import logging

import pytest
from jupyter_server.utils import ensure_async
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook

import jupytext
from jupytext.compare import compare_cells, notebook_model


@pytest.fixture(params=["sync", "async"])
def cm_from_fs_meta_manager(tmpdir, request):
    try:
        from jupyterfs.metamanager import MetaManager, SyncMetaManager
    except ImportError:
        pytest.skip("jupyterfs is not available")

    if request.param == "sync":
        cm_class = jupytext.build_sync_jupytext_contents_manager_class(SyncMetaManager)
    else:
        cm_class = jupytext.build_async_jupytext_contents_manager_class(MetaManager)

    logger = logging.getLogger("jupyter-fs")
    cm = cm_class(parent=None, log=logger)
    cm.initResource(
        {
            "url": f"osfs://{tmpdir}",
        }
    )
    return cm


@pytest.mark.asyncio
async def test_jupytext_jupyter_fs_metamanager(cm_from_fs_meta_manager):
    """Test the basic get/save functions of Jupytext with a fs manager
    https://github.com/mwouts/jupytext/issues/618"""
    cm = cm_from_fs_meta_manager
    # the hash that corresponds to the osfs
    osfs = [h for h in cm._managers if h != ""][0]

    # save a few files
    text = "some text\n"
    await ensure_async(
        cm.save(dict(type="file", content=text, format="text"), path=osfs + ":text.md")
    )
    nb = new_notebook(
        cells=[new_markdown_cell("A markdown cell"), new_code_cell("1 + 1")]
    )
    await ensure_async(cm.save(notebook_model(nb), osfs + ":notebook.ipynb"))
    await ensure_async(cm.save(notebook_model(nb), osfs + ":text_notebook.md"))

    # list the directory
    directory = await ensure_async(cm.get(osfs + ":/"))
    assert {file["name"] for file in directory["content"]} == {
        "text.md",
        "text_notebook.md",
        "notebook.ipynb",
    }

    # get the files
    model = await ensure_async(cm.get(osfs + ":/text.md", type="file"))
    assert model["type"] == "file"
    assert model["content"] == text

    model = await ensure_async(cm.get(osfs + ":text.md", type="notebook"))
    assert model["type"] == "notebook"
    # We only compare the cells, as kernelspecs are added to the notebook metadata
    compare_cells(
        model["content"].cells, [new_markdown_cell(text.strip())], compare_ids=False
    )

    for nb_file in ["notebook.ipynb", "text_notebook.md"]:
        model = await ensure_async(cm.get(osfs + ":" + nb_file))
        assert model["type"] == "notebook"
        actual_cells = model["content"].cells

        # saving adds 'trusted=True' to the code cell metadata
        for cell in actual_cells:
            cell.metadata = {}
        compare_cells(actual_cells, nb.cells, compare_ids=False)


@pytest.mark.asyncio
async def test_config_jupytext_jupyter_fs_meta_manager(tmpdir, cm_from_fs_meta_manager):
    """Test the configuration of Jupytext with a fs manager"""
    tmpdir.join("jupytext.toml").write('formats = "ipynb,py"')
    cm = cm_from_fs_meta_manager
    # the hash that corresponds to the osfs
    osfs = [h for h in cm._managers if h != ""][0]

    # save a few files
    nb = new_notebook()
    await ensure_async(
        cm.save(
            dict(type="file", content="text", format="text"), path=osfs + ":text.md"
        )
    )
    await ensure_async(cm.save(notebook_model(nb), osfs + ":script.py"))
    await ensure_async(cm.save(notebook_model(nb), osfs + ":text_notebook.md"))
    await ensure_async(cm.save(notebook_model(nb), osfs + ":notebook.ipynb"))

    # list the directory
    directory = await ensure_async(cm.get(osfs + ":/"))
    assert {file["name"] for file in directory["content"]} == {
        "jupytext.toml",
        "text.md",
        "text_notebook.md",
        "notebook.ipynb",
        "notebook.py",
        "script.py",
        "script.ipynb",
    }



from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook

import jupytext
from jupytext.compare import compare, compare_notebooks


def test_read_simple_code(
    no_jupytext_version_number,
    ml="""(* %% *)
let sum x y = x + y
""",
    nb=new_notebook(
        cells=[
            new_code_cell("""let sum x y = x + y"""),
        ]
    ),
):
    nb2 = jupytext.reads(ml, "ml:percent")
    compare_notebooks(nb2, nb)
    ml2 = jupytext.writes(nb, "ml:percent")
    compare(ml2, ml)


def test_read_simple_markdown(
    no_jupytext_version_number,
    ml="""(* %% [markdown] *)
(* # Example of an OCaml notebook *)
""",
    nb=new_notebook(
        cells=[
            new_markdown_cell("# Example of an OCaml notebook"),
        ]
    ),
):
    nb2 = jupytext.reads(ml, "ml:percent")
    compare_notebooks(nb2, nb)
    ml2 = jupytext.writes(nb, "ml:percent")
    compare(ml2, ml)



import pytest
from git.exc import HookExecutionError
from nbformat.v4.nbbase import new_markdown_cell
from pre_commit.main import main as pre_commit

from jupytext import TextFileContentsManager, read


def test_pre_commit_hook_sync_with_config(
    tmpdir,
    cwd_tmpdir,
    tmp_repo,
    jupytext_repo_root,
    jupytext_repo_rev,
    python_notebook,
):
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [--sync]
"""
    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)

    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks", "-f"])

    tmpdir.join("jupytext.toml").write('formats = "ipynb,py:percent"\n')

    # create a test notebook and save it in Jupyter
    nb = python_notebook
    cm = TextFileContentsManager()
    cm.root_dir = str(tmpdir)
    cm.save(dict(type="notebook", content=nb), "test.ipynb")

    # Assert that "text_representation" is in the Jupytext metadata #900
    assert "text_representation" in tmpdir.join("test.py").read()
    # But not in the ipynb notebook
    assert "text_representation" not in tmpdir.join("test.ipynb").read()

    # try to commit it, should fail as the py version hasn't been added
    tmp_repo.git.add("test.ipynb")
    with pytest.raises(
        HookExecutionError,
        match="git add test.py",
    ):
        tmp_repo.index.commit("failing")

    # add the py file, now the commit will succeed
    tmp_repo.git.add("test.py")
    tmp_repo.index.commit("passing")
    assert "test.ipynb" in tmp_repo.tree()
    assert "test.py" in tmp_repo.tree()

    # The text representation metadata is still in the py file
    assert "text_representation" in tmpdir.join("test.py").read()
    # But not in the ipynb notebook
    assert "text_representation" not in tmpdir.join("test.ipynb").read()

    # modify the ipynb file in Jupyter
    # Reload the notebook
    nb = cm.get("test.ipynb")["content"]
    nb.cells.append(new_markdown_cell("A new cell"))
    cm.save(dict(type="notebook", content=nb), "test.ipynb")

    # The text representation metadata is in the py file
    assert "text_representation" in tmpdir.join("test.py").read()
    # But not in the ipynb notebook
    assert "text_representation" not in tmpdir.join("test.ipynb").read()

    # the text file has been updated (and the ipynb file as well)
    assert "A new cell" in tmpdir.join("test.py").read()
    nb = read("test.ipynb")
    assert len(nb.cells) == 2

    # add both files, the commit will succeed
    tmp_repo.git.add("test.ipynb")
    tmp_repo.git.add("test.py")
    tmp_repo.index.commit("passing")



import pytest
from git.exc import HookExecutionError
from pre_commit.main import main as pre_commit

from jupytext import read, write


def test_pre_commit_hook_sync_black_nbstripout(
    tmpdir,
    cwd_tmpdir,
    tmp_repo,
    jupytext_repo_root,
    jupytext_repo_rev,
    notebook_with_outputs,
):
    """Here we sync the ipynb notebook with a py:percent file and also apply black and nbstripout."""
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [--sync, --pipe, black]
    additional_dependencies:
    - black==22.3.0  # Matches hook

- repo: https://github.com/psf/black
  rev: 22.3.0
  hooks:
  - id: black

- repo: https://github.com/kynan/nbstripout
  rev: 0.5.0
  hooks:
  - id: nbstripout
"""
    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)

    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks", "-f"])

    tmpdir.join("jupytext.toml").write('formats = "ipynb,py:percent"')
    tmp_repo.git.add("jupytext.toml")
    tmp_repo.index.commit("pair notebooks")

    # write a test notebook
    write(notebook_with_outputs, "test.ipynb")

    # try to commit it, should fail because
    # 1. the py version hasn't been added
    # 2. the first cell is '1+1' which is not black compliant
    # 3. the notebook has outputs
    tmp_repo.git.add("test.ipynb")
    with pytest.raises(
        HookExecutionError,
        match="files were modified by this hook",
    ):
        tmp_repo.index.commit("failing")

    # Add the two files
    tmp_repo.git.add("test.ipynb")
    tmp_repo.git.add("test.py")

    # now the commit will succeed
    tmp_repo.index.commit("passing")
    assert "test.ipynb" in tmp_repo.tree()
    assert "test.py" in tmp_repo.tree()

    # the first cell was reformatted
    nb = read("test.ipynb")
    assert nb.cells[0].source == "1 + 1"

    # the ipynb file has no outputs
    assert not nb.cells[0].outputs



import nbformat
import pytest
from packaging import version

import jupytext


@pytest.mark.skipif(
    version.parse(nbformat.__version__) <= version.parse("5.7"),
    reason="normalize is not available",
)
def test_sample_notebooks_are_normalized(any_nb_file):
    if "myst/" in any_nb_file and not jupytext.myst.is_myst_available():
        pytest.skip("myst_parser not found")
    nb = jupytext.read(any_nb_file)

    changes, normalized_nb = nbformat.validator.normalize(nb)
    nbformat.validate(normalized_nb)

    if changes:  # pragma: no cover
        with open(any_nb_file, "w") as fp:
            jupytext.write(normalized_nb, fp)

        assert not changes

    nbformat.validate(nb)



import os
import unittest.mock as mock

import pytest

import jupytext
from jupytext.cli import jupytext as jupytext_cli
from jupytext.compare import compare
from jupytext.formats import (
    long_form_multiple_formats,
    long_form_one_format,
    short_form_multiple_formats,
)
from jupytext.paired_paths import InconsistentPath, base_path, full_path, paired_paths


def test_simple_pair():
    formats = long_form_multiple_formats("ipynb,py")
    expected_paths = ["notebook.ipynb", "notebook.py"]
    compare(
        paired_paths("notebook.ipynb", "ipynb", formats),
        list(zip(expected_paths, formats)),
    )
    compare(
        paired_paths("notebook.py", "py", formats), list(zip(expected_paths, formats))
    )


def test_base_path():
    fmt = long_form_one_format("dir/prefix_/ipynb")
    assert base_path("dir/prefix_NAME.ipynb", fmt) == "NAME"
    with pytest.raises(InconsistentPath):
        base_path("dir/incorrect_prefix_NAME.ipynb", fmt)


def test_base_path_dotdot():
    fmt = long_form_one_format("../scripts//py")
    assert base_path("scripts/test.py", fmt=fmt) == "scripts/test"


def test_full_path_dotdot():
    fmt = long_form_one_format("../scripts//py")
    assert full_path("scripts/test", fmt=fmt) == "scripts/test.py"


def test_base_path_in_tree_from_root():
    fmt = long_form_one_format("scripts///py")
    assert base_path("scripts/subfolder/test.py", fmt=fmt) == "//subfolder/test"
    assert base_path("/scripts/subfolder/test.py", fmt=fmt) == "///subfolder/test"


def test_base_path_in_tree_from_non_root():
    fmt = long_form_one_format("scripts///py")
    assert (
        base_path("/parent_folder/scripts/subfolder/test.py", fmt=fmt)
        == "/parent_folder///subfolder/test"
    )


def test_base_path_in_tree_from_non_root_no_subfolder():
    nb_file = "/parent/notebooks/wrap_markdown.ipynb"
    formats = "notebooks///ipynb,scripts///py:percent"
    fmt = "notebooks///ipynb"
    assert base_path(nb_file, fmt) == "/parent///wrap_markdown"
    paired_paths(nb_file, fmt, formats)


def test_full_path_in_tree_from_root():
    fmt = long_form_one_format("notebooks///ipynb")
    assert full_path("//subfolder/test", fmt=fmt) == "notebooks/subfolder/test.ipynb"
    assert full_path("///subfolder/test", fmt=fmt) == "/notebooks/subfolder/test.ipynb"


def test_full_path_in_tree_from_root_no_subfolder():
    fmt = long_form_one_format("notebooks///ipynb")
    assert full_path("//test", fmt=fmt) == "notebooks/test.ipynb"
    assert full_path("///test", fmt=fmt) == "/notebooks/test.ipynb"


def test_full_path_in_tree_from_non_root():
    fmt = long_form_one_format("notebooks///ipynb")
    assert (
        full_path("/parent_folder///subfolder/test", fmt=fmt)
        == "/parent_folder/notebooks/subfolder/test.ipynb"
    )


def test_paired_paths_windows():
    nb_file = "C:\\Users\\notebooks\\notebooks\\subfolder\\nb.ipynb"
    formats = "notebooks///ipynb,scripts///py"
    with mock.patch("os.path.sep", "\\"):
        assert (
            base_path(nb_file, "notebooks///ipynb")
            == "C:\\Users\\notebooks\\//subfolder\\nb"
        )
        paired_paths(nb_file, "notebooks///ipynb", formats)


def test_paired_paths_windows_no_subfolder():
    nb_file = "C:\\Users\\notebooks\\notebooks\\nb.ipynb"
    formats = "notebooks///ipynb,scripts///py"
    with mock.patch("os.path.sep", "\\"):
        assert base_path(nb_file, "notebooks///ipynb") == "C:\\Users\\notebooks\\//nb"
        paired_paths(nb_file, "notebooks///ipynb", formats)


@pytest.mark.parametrize("os_path_sep", ["\\", "/"])
def test_paired_path_dotdot_564(os_path_sep):
    main_path = os_path_sep.join(
        ["examples", "tutorials", "colabs", "rigid_object_tutorial.ipynb"]
    )
    formats = "../nb_python//py:percent,../colabs//ipynb"
    with mock.patch("os.path.sep", os_path_sep):
        assert base_path(
            main_path, None, long_form_multiple_formats(formats)
        ) == os_path_sep.join(
            ["examples", "tutorials", "colabs", "rigid_object_tutorial"]
        )
        paired_paths(main_path, "ipynb", formats)


def test_path_in_tree_limited_to_config_dir(tmpdir):
    root_nb_dir = tmpdir.mkdir("notebooks")
    nb_dir = root_nb_dir.mkdir("notebooks")
    src_dir = root_nb_dir.mkdir("scripts")
    other_dir = root_nb_dir.mkdir("other")

    formats = "notebooks///ipynb,scripts///py"
    fmt = long_form_multiple_formats(formats)[0]

    # Notebook in nested 'notebook' dir is paired
    notebook_in_nb_dir = nb_dir.join("subfolder").join("nb.ipynb")

    assert {
        path for (path, _) in paired_paths(str(notebook_in_nb_dir), fmt, formats)
    } == {str(notebook_in_nb_dir), str(src_dir.join("subfolder").join("nb.py"))}

    # Notebook in base 'notebook' dir is paired if no config file is found
    notebook_in_other_dir = other_dir.mkdir("subfolder").join("nb.ipynb")
    assert {
        path for (path, _) in paired_paths(str(notebook_in_other_dir), fmt, formats)
    } == {
        str(notebook_in_other_dir),
        str(tmpdir.join("scripts").join("other").join("subfolder").join("nb.py")),
    }

    # When a config file exists
    root_nb_dir.join("jupytext.toml").write("\n")

    # Notebook in nested 'notebook' dir is still paired
    assert {
        path for (path, _) in paired_paths(str(notebook_in_nb_dir), fmt, formats)
    } == {str(notebook_in_nb_dir), str(src_dir.join("subfolder").join("nb.py"))}

    # But the notebook in base 'notebook' dir is not paired any more
    alert = (
        "Notebook directory '/other/subfolder' does not match prefix root 'notebooks'"
        if os.path.sep == "/"
        # we escape twice the backslash because pytest.raises matches it as a regular expression
        else "Notebook directory '\\\\other\\\\subfolder' does not match prefix root 'notebooks'"
    )
    with pytest.raises(InconsistentPath, match=alert):
        paired_paths(str(notebook_in_other_dir), fmt, formats)


def test_many_and_suffix():
    formats = long_form_multiple_formats("ipynb,.pct.py,_lgt.py")
    expected_paths = ["notebook.ipynb", "notebook.pct.py", "notebook_lgt.py"]
    for fmt, path in zip(formats, expected_paths):
        compare(paired_paths(path, fmt, formats), list(zip(expected_paths, formats)))

    with pytest.raises(InconsistentPath):
        paired_paths("wrong_suffix.py", "py", formats)


def test_prefix_and_suffix():
    short_formats = (
        "notebook_folder/notebook_prefix_/_notebook_suffix.ipynb,"
        "script_folder//_in_percent_format.py:percent,"
        "script_folder//_in_light_format.py"
    )

    formats = long_form_multiple_formats(short_formats)
    assert short_form_multiple_formats(formats) == short_formats

    expected_paths = [
        "parent/notebook_folder/notebook_prefix_NOTEBOOK_NAME_notebook_suffix.ipynb",
        "parent/script_folder/NOTEBOOK_NAME_in_percent_format.py",
        "parent/script_folder/NOTEBOOK_NAME_in_light_format.py",
    ]
    for fmt, path in zip(formats, expected_paths):
        compare(paired_paths(path, fmt, formats), list(zip(expected_paths, formats)))

    # without the parent folder
    expected_paths = [path[7:] for path in expected_paths]
    for fmt, path in zip(formats, expected_paths):
        compare(paired_paths(path, fmt, formats), list(zip(expected_paths, formats)))

    # Not the expected parent folder
    with pytest.raises(InconsistentPath):
        paired_paths(
            "script_folder_incorrect/NOTEBOOK_NAME_in_percent_format.py",
            formats[1],
            formats,
        )

    # Not the expected suffix
    with pytest.raises(InconsistentPath):
        paired_paths(
            "parent/script_folder/NOTEBOOK_NAME_in_LIGHT_format.py", formats[2], formats
        )

    # Not the expected extension
    with pytest.raises(InconsistentPath):
        paired_paths(
            "notebook_folder/notebook_prefix_NOTEBOOK_NAME_notebook_suffix.py",
            formats[0],
            formats,
        )


def test_prefix_on_root_174():
    short_formats = "ipynb,python//py:percent"

    formats = long_form_multiple_formats(short_formats)
    assert short_form_multiple_formats(formats) == short_formats

    expected_paths = ["Untitled.ipynb", "python/Untitled.py"]
    for fmt, path in zip(formats, expected_paths):
        compare(paired_paths(path, fmt, formats), list(zip(expected_paths, formats)))


def test_duplicated_paths():
    formats = long_form_multiple_formats("ipynb,py:percent,py:light")
    with pytest.raises(InconsistentPath):
        paired_paths("notebook.ipynb", "ipynb", formats)


@pytest.mark.asyncio
async def test_cm_paired_paths(cm):
    cm.paired_notebooks = dict()

    three = "ipynb,py,md"
    cm.update_paired_notebooks("nb.ipynb", three)
    assert cm.paired_notebooks == {
        "nb." + fmt: (fmt, three) for fmt in three.split(",")
    }

    two = "ipynb,Rmd"
    cm.update_paired_notebooks("nb.ipynb", two)
    assert cm.paired_notebooks == {"nb." + fmt: (fmt, two) for fmt in two.split(",")}

    one = "ipynb"
    cm.update_paired_notebooks("nb.ipynb", one)
    assert cm.paired_notebooks == {}

    zero = ""
    cm.update_paired_notebooks("nb.ipynb", zero)
    assert cm.paired_notebooks == {}


def test_paired_path_with_prefix(
    nb_file="scripts/test.py",
    fmt={"extension": ".py", "format_name": "percent"},
    formats=[
        {"extension": ".ipynb"},
        {"prefix": "scripts/", "format_name": "percent", "extension": ".py"},
    ],
):
    assert paired_paths(nb_file, fmt, formats) == [
        ("test.ipynb", {"extension": ".ipynb"}),
        (
            "scripts/test.py",
            {"prefix": "scripts/", "format_name": "percent", "extension": ".py"},
        ),
    ]


def test_paired_notebook_ipynb_root_scripts_in_folder_806(
    tmpdir, cwd_tmpdir, python_notebook
):
    """In this test we pair a notebook with a script in a subfolder, and then do some
    natural operations like delete/recreate one of the paired files"""
    # Save sample notebook
    test_ipynb = tmpdir / "test.ipynb"
    jupytext.write(python_notebook, str(test_ipynb))

    # Pair the notebook to a script in a subfolder
    jupytext_cli(["--set-formats", "ipynb,scripts//py:percent", "test.ipynb"])
    assert (tmpdir / "scripts" / "test.py").exists()

    # Delete and then recreate the ipynb notebook using --to notebook
    test_ipynb.remove()
    jupytext_cli(
        [
            "--to",
            "notebook",
            "--output",
            "test.ipynb",
            "scripts/test.py",
        ]
    )
    assert test_ipynb.exists()

    # Delete and then recreate the ipynb notebook using --sync
    test_ipynb.remove()
    jupytext_cli(
        [
            "--sync",
            "scripts/test.py",
        ]
    )
    assert test_ipynb.exists()



import os
import stat
import unittest.mock as mock

import pytest
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook

from jupytext import read, write
from jupytext.cli import jupytext, system
from jupytext.compare import compare_cells, compare_notebooks


def git_in_tmpdir(tmpdir):
    """Return a function that will execute git instruction in the desired directory"""

    def git(*args):
        out = system("git", *args, cwd=str(tmpdir))
        print(out)
        return out

    git("init", "--initial-branch", "main")
    git("status")
    git("config", "user.name", "jupytext-test-cli")
    git("config", "user.email", "jupytext@tests.com")

    return git


def system_in_tmpdir(tmpdir):
    """Return a function that will execute system commands in the desired directory"""

    def system_(*args):
        return system(*args, cwd=str(tmpdir))

    return system_


def test_pre_commit_hook(tmpdir):
    tmp_ipynb = str(tmpdir.join("nb with spaces.ipynb"))
    tmp_py = str(tmpdir.join("nb with spaces.py"))
    nb = new_notebook(cells=[])

    git = git_in_tmpdir(tmpdir)
    hook = str(tmpdir.join(".git/hooks/pre-commit"))
    with open(hook, "w") as fp:
        fp.write("#!/bin/sh\n" "jupytext --to py:percent --pre-commit\n")

    st = os.stat(hook)
    os.chmod(hook, st.st_mode | stat.S_IEXEC)

    write(nb, tmp_ipynb)
    assert os.path.isfile(tmp_ipynb)
    assert not os.path.isfile(tmp_py)

    git("add", "nb with spaces.ipynb")
    git("status")
    git("commit", "-m", "created")
    git("status")

    assert "nb with spaces.py" in git("ls-tree", "-r", "main", "--name-only")
    assert os.path.isfile(tmp_py)


def test_sync_with_pre_commit_hook(tmpdir):
    # Init git and create a pre-commit hook
    git = git_in_tmpdir(tmpdir)
    hook = str(tmpdir.join(".git/hooks/pre-commit"))
    with open(hook, "w") as fp:
        fp.write("#!/bin/sh\n" "jupytext --sync --pre-commit\n")

    st = os.stat(hook)
    os.chmod(hook, st.st_mode | stat.S_IEXEC)

    # Create a notebook that is not paired
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_md = str(tmpdir.join("notebook.md"))
    nb = new_notebook(cells=[new_markdown_cell("A short notebook")])
    write(nb, tmp_ipynb)
    assert os.path.isfile(tmp_ipynb)
    assert not os.path.isfile(tmp_md)

    git("add", "notebook.ipynb")
    git("status")
    git("commit", "-m", "created")
    git("status")

    assert "notebook.ipynb" in git("ls-tree", "-r", "main", "--name-only")
    assert "notebook.md" not in git("ls-tree", "-r", "main", "--name-only")

    assert os.path.isfile(tmp_ipynb)
    assert not os.path.exists(tmp_md)

    # Pair the notebook
    jupytext(["--set-formats", "ipynb,md", tmp_ipynb])

    # Remove the md file (it will be regenerated by the pre-commit hook)
    os.remove(tmp_md)

    # Commit the ipynb file
    git("add", "notebook.ipynb")
    git("status")
    git("commit", "-m", "paired")
    git("status")

    # The pre-commit script should have created and committed the md file
    assert "notebook.ipynb" in git("ls-tree", "-r", "main", "--name-only")
    assert "notebook.md" in git("ls-tree", "-r", "main", "--name-only")
    assert os.path.isfile(tmp_md)
    nb_md = read(tmp_md)
    compare_notebooks(nb_md, nb)

    # Edit the md file
    with open(tmp_md) as fp:
        md_text = fp.read()

    with open(tmp_md, "w") as fp:
        fp.write(md_text.replace("A short notebook", "Notebook was edited"))

    # commit the md file
    git("add", "notebook.md")
    git("status")
    git("commit", "-m", "edited md")
    git("status")

    # The pre-commit script should have sync and committed the ipynb file
    assert "notebook.ipynb" in git("ls-tree", "-r", "main", "--name-only")
    assert "notebook.md" in git("ls-tree", "-r", "main", "--name-only")

    nb = read(tmp_ipynb)
    compare_cells(
        nb.cells, [new_markdown_cell("Notebook was edited")], compare_ids=False
    )

    # create and commit a jpg file
    tmp_jpg = str(tmpdir.join("image.jpg"))
    with open(tmp_jpg, "wb") as fp:
        fp.write(b"")
    git("add", "image.jpg")
    git("commit", "-m", "added image")


def test_pre_commit_hook_in_subfolder(tmpdir):
    tmp_ipynb = str(tmpdir.join("nb with spaces.ipynb"))
    tmp_py = str(tmpdir.join("python", "nb with spaces.py"))
    nb = new_notebook(cells=[])

    git = git_in_tmpdir(tmpdir)
    hook = str(tmpdir.join(".git/hooks/pre-commit"))
    with open(hook, "w") as fp:
        fp.write(
            "#!/bin/sh\n" "jupytext --from ipynb --to python//py:percent --pre-commit\n"
        )

    st = os.stat(hook)
    os.chmod(hook, st.st_mode | stat.S_IEXEC)

    write(nb, tmp_ipynb)
    assert os.path.isfile(tmp_ipynb)
    assert not os.path.isfile(tmp_py)

    git("add", "nb with spaces.ipynb")
    git("status")
    git("commit", "-m", "created")
    git("status")

    assert "nb with spaces.py" in git("ls-tree", "-r", "main", "--name-only")
    assert os.path.isfile(tmp_py)


def test_pre_commit_hook_py_to_ipynb_and_md(tmpdir):
    tmp_ipynb = str(tmpdir.join("nb with spaces.ipynb"))
    tmp_py = str(tmpdir.join("nb with spaces.py"))
    tmp_md = str(tmpdir.join("nb with spaces.md"))
    nb = new_notebook(cells=[])

    git = git_in_tmpdir(tmpdir)
    hook = str(tmpdir.join(".git/hooks/pre-commit"))
    with open(hook, "w") as fp:
        fp.write(
            "#!/bin/sh\n"
            "jupytext --from py:percent --to ipynb --pre-commit\n"
            "jupytext --from py:percent --to md --pre-commit\n"
        )

    st = os.stat(hook)
    os.chmod(hook, st.st_mode | stat.S_IEXEC)

    write(nb, tmp_py)
    assert os.path.isfile(tmp_py)
    assert not os.path.isfile(tmp_ipynb)
    assert not os.path.isfile(tmp_md)

    git("add", "nb with spaces.py")
    git("status")
    git("commit", "-m", "created")
    git("status")

    assert "nb with spaces.ipynb" in git("ls-tree", "-r", "main", "--name-only")
    assert "nb with spaces.md" in git("ls-tree", "-r", "main", "--name-only")

    assert os.path.isfile(tmp_ipynb)
    assert os.path.isfile(tmp_md)


@pytest.mark.requires_black
@pytest.mark.requires_flake8
def test_pre_commit_hook_sync_black_flake8(tmpdir, python_notebook):
    # Load real notebook metadata to get the 'auto' extension in --pipe-fmt to work
    metadata = python_notebook.metadata

    git = git_in_tmpdir(tmpdir)
    hook = str(tmpdir.join(".git/hooks/pre-commit"))
    with open(hook, "w") as fp:
        fp.write(
            "#!/bin/sh\n"
            "# Pair ipynb notebooks to a python file, reformat content with black, and run flake8\n"
            "# Note: this hook only acts on ipynb files. When pulling, run 'jupytext --sync' to "
            "update the ipynb file.\n"
            "jupytext --pre-commit --from ipynb --set-formats ipynb,py --pipe black --check flake8\n"
        )

    st = os.stat(hook)
    os.chmod(hook, st.st_mode | stat.S_IEXEC)

    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))
    nb = new_notebook(cells=[new_code_cell(source="1+    1")], metadata=metadata)

    write(nb, tmp_ipynb)
    git("add", "notebook.ipynb")
    git("status")
    git("commit", "-m", "created")
    git("status")
    assert os.path.isfile(tmp_py)
    assert os.path.isfile(tmp_ipynb)
    with open(tmp_py) as fp:
        assert fp.read().splitlines()[-1] == "1 + 1"

    nb = new_notebook(
        cells=[new_code_cell(source='"""trailing   \nwhitespace"""    ')],
        metadata=metadata,
    )
    write(nb, tmp_ipynb)
    git("add", "notebook.ipynb")
    git("status")
    with open(tmp_py) as fp:
        assert fp.read().splitlines()[-1] == "1 + 1"


@pytest.mark.requires_flake8
def test_pre_commit_hook_sync_flake8(tmpdir, python_notebook):
    """This test was extracted from test_pre_commit_hook_sync_black_flake8 to make sure that the
    pre-commit stops the commit when flake8 fails"""
    metadata = python_notebook.metadata

    git = git_in_tmpdir(tmpdir)
    hook = str(tmpdir.join(".git/hooks/pre-commit"))
    with open(hook, "w") as fp:
        fp.write(
            "#!/bin/sh\n"
            "# Pair ipynb notebooks to a python file, and run flake8\n"
            "# Note: this hook only acts on ipynb files. When pulling, run 'jupytext --sync' to "
            "update the ipynb file.\n"
            "jupytext --pre-commit --from ipynb --set-formats ipynb,py --check flake8\n"
        )

    st = os.stat(hook)
    os.chmod(hook, st.st_mode | stat.S_IEXEC)

    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    nb = new_notebook(
        cells=[new_code_cell(source='"""trailing   \nwhitespace"""    ')],
        metadata=metadata,
    )
    write(nb, tmp_ipynb)
    git("add", "notebook.ipynb")
    git("status")
    with pytest.raises(SystemExit):  # not flake8
        git("commit", "-m", "created")


@pytest.mark.filterwarnings("ignore:The --pre-commit argument is deprecated")
def test_manual_call_of_pre_commit_hook(tmpdir):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))
    nb = new_notebook(cells=[])
    os.chdir(str(tmpdir))

    git = git_in_tmpdir(tmpdir)

    def hook():
        with mock.patch("jupytext.cli.system", system_in_tmpdir(tmpdir)):
            jupytext(["--to", "py", "--pre-commit"])

    write(nb, tmp_ipynb)
    assert os.path.isfile(tmp_ipynb)
    assert not os.path.isfile(tmp_py)

    git("add", "notebook.ipynb")
    git("status")
    hook()
    git("commit", "-m", "created")
    git("status")

    assert "notebook.py" in git("ls-tree", "-r", "main", "--name-only")
    assert os.path.isfile(tmp_py)


def test_pre_commit_hook_with_subfolders_issue_506(tmpdir):
    """I have the following directory structure, where the nb/test.ipynb is paired with the py/test.py.

    ├── nb
    │   └── test.ipynb
    └── py
        └── test.py
    """

    nb_file = tmpdir.mkdir("nb").join("test.ipynb")
    py_file = tmpdir.mkdir("py").join("test.py")

    """The notebook and Python file are paired with "jupytext": {"formats": "py//py,nb//ipynb"}.
        (using jupytext --set-formats py//py,nb//ipynb nb/test.ipynb)"""
    write(
        new_notebook(
            cells=[new_markdown_cell("A Markdown cell")],
            metadata={"jupytext": {"formats": "py//py,nb//ipynb"}},
        ),
        str(py_file),
    )

    """This works fine when syncing with jupytext --sync nb/test.ipynb
    but when syncing with jupytext --sync --pre-commit I get the following exception: (...)"""
    git = git_in_tmpdir(tmpdir)
    hook = str(tmpdir.join(".git/hooks/pre-commit"))
    with open(hook, "w") as fp:
        fp.write("#!/bin/sh\n" "jupytext --sync --pre-commit\n")

    st = os.stat(hook)
    os.chmod(hook, st.st_mode | stat.S_IEXEC)

    assert not os.path.isfile(str(nb_file))

    git("add", "py/test.py")
    git("status")
    git("commit", "-m", "notebook created")
    git("status")

    assert os.path.isfile(str(nb_file))
    assert read(str(nb_file)).cells[0].source == "A Markdown cell"


@pytest.mark.requires_pandoc
def test_wrap_markdown_cell(tmpdir):
    """Use a pre-commit hook to sync a notebook to a script paired in a tree, and reformat
    the markdown cells using pandoc"""

    tmpdir.join("jupytext.toml").write(
        """# By default, the notebooks in this repository are in the notebooks subfolder
    # and they are paired to scripts in the script subfolder.
    formats = "notebooks///ipynb,scripts///py:percent"
    """
    )

    git = git_in_tmpdir(tmpdir)
    hook = str(tmpdir.join(".git/hooks/pre-commit"))
    with open(hook, "w") as fp:
        fp.write(
            "#!/bin/sh\n"
            "jupytext --pre-commit --sync --pipe-fmt ipynb --pipe \\\n"
            "    'pandoc --from ipynb --to ipynb --markdown-headings=atx'\n"
        )

    st = os.stat(hook)
    os.chmod(hook, st.st_mode | stat.S_IEXEC)

    nb_file = tmpdir.mkdir("notebooks").mkdir("subfolder").join("wrap_markdown.ipynb")
    long_text = "This is a " + ("very " * 24) + "long sentence."
    nb = new_notebook(cells=[new_markdown_cell(long_text)])
    write(nb, str(nb_file))

    nb = read(str(nb_file))
    assert nb.cells[0].source == long_text

    git("add", str(nb_file))
    git("commit", "-m", "'notebook with long cells'")

    py_text = tmpdir.join("scripts").join("subfolder").join("wrap_markdown.py").read()
    assert "This is a very very" in py_text
    for line in py_text.splitlines():
        assert len(line) <= 79

    nb = read(nb_file, as_version=4)
    text = nb.cells[0].source
    assert len(text.splitlines()) >= 2
    assert text != long_text



"""Here we generate mirror representation of py, Rmd and ipynb files
as py or ipynb, and make sure that these representations minimally
change on new releases.
"""

import os
import re

import pytest
from nbformat.v4.nbbase import new_notebook

import jupytext
from jupytext.compare import (
    assert_conversion_same_as_mirror,
    compare_notebooks,
    create_mirror_file_if_missing,
)
from jupytext.formats import auto_ext_from_metadata
from jupytext.languages import _SCRIPT_EXTENSIONS


def test_create_mirror_file_if_missing(tmpdir, no_jupytext_version_number):
    py_file = str(tmpdir.join("notebook.py"))
    assert not os.path.isfile(py_file)
    create_mirror_file_if_missing(py_file, new_notebook(), "py")
    assert os.path.isfile(py_file)


"""---------------------------------------------------------------------------------

Part I: ipynb -> fmt -> ipynb

---------------------------------------------------------------------------------"""


def test_ipynb_to_percent(ipynb_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_file, "auto:percent", "ipynb_to_percent")


def test_ipynb_to_hydrogen(ipynb_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_file, "auto:hydrogen", "ipynb_to_hydrogen")


def test_ipynb_to_light(ipynb_to_light, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_to_light, "auto:light", "ipynb_to_script")


def test_ipynb_to_md(ipynb_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_file, "md", "ipynb_to_md")


def test_ipynb_to_Rmd(ipynb_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_file, "Rmd", "ipynb_to_Rmd")


@pytest.mark.requires_myst
def test_ipynb_to_myst(ipynb_file, no_jupytext_version_number):
    if re.match(
        r".*(html-demo|julia_functional_geometry|xcpp_by_quantstack).*", ipynb_file
    ):
        pytest.skip()
    assert_conversion_same_as_mirror(ipynb_file, "md:myst", "ipynb_to_myst")


"""---------------------------------------------------------------------------------

Part II: text -> ipynb -> text

---------------------------------------------------------------------------------"""


def test_script_to_ipynb(script_to_ipynb, no_jupytext_version_number):
    assert_conversion_same_as_mirror(script_to_ipynb, "ipynb:light", "script_to_ipynb")


def test_percent_to_ipynb(percent_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(percent_file, "ipynb:percent", "script_to_ipynb")


def test_hydrogen_to_ipynb(hydrogen_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(hydrogen_file, "ipynb:hydrogen", "script_to_ipynb")


def test_spin_to_ipynb(r_spin_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(r_spin_file, "ipynb:spin", "script_to_ipynb")


def test_md_to_ipynb(md_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(md_file, "ipynb", "md_to_ipynb")


@pytest.mark.requires_myst
def test_myst_file_has_myst_format(myst_file):
    with open(myst_file) as f:
        text = f.read()
    fmt = jupytext.guess_format(text, ".md")
    assert fmt == ("myst", {})


@pytest.mark.requires_myst
def test_myst_to_ipynb(myst_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(myst_file, "ipynb:myst", "myst_to_ipynb")


"""---------------------------------------------------------------------------------

Part III: More specific round trip tests

---------------------------------------------------------------------------------"""


def test_ipynb_to_percent_to_light(ipynb_file):
    nb = jupytext.read(ipynb_file)
    pct = jupytext.writes(nb, "auto:percent")
    auto_ext = auto_ext_from_metadata(nb.metadata)
    comment = _SCRIPT_EXTENSIONS[auto_ext]["comment"]
    lgt = (
        pct.replace(comment + " %%\n", comment + " +\n")
        .replace(comment + " %% ", comment + " + ")
        .replace(
            comment + "       format_name: percent",
            comment + "       format_name: light",
        )
    )
    nb2 = jupytext.reads(lgt, auto_ext)
    compare_notebooks(nb2, nb)


def test_ipynb_to_python_vim(ipynb_py_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(
        ipynb_py_file,
        {"extension": ".py", "cell_markers": "{{{,}}}"},
        "ipynb_to_script_vim_folding_markers",
    )


def test_ipynb_to_python_vscode(ipynb_py_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(
        ipynb_py_file,
        {"extension": ".py", "cell_markers": "region,endregion"},
        "ipynb_to_script_vscode_folding_markers",
    )


def test_ipynb_to_r_light(ipynb_R_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_R_file, ".low.r:light", "ipynb_to_script")


def test_ipynb_to_r_percent(ipynb_R_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_R_file, ".low.r:percent", "ipynb_to_percent")


def test_ipynb_to_R_spin(ipynb_R_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_R_file, "R:spin", "ipynb_to_spin")


def test_ipynb_to_r_spin(ipynb_R_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_R_file, ".low.r:spin", "ipynb_to_spin")


@pytest.mark.parametrize("extension", ("ss", "scm"))
def test_ipynb_to_scheme_light(
    ipynb_scheme_file, extension, no_jupytext_version_number
):
    assert_conversion_same_as_mirror(
        ipynb_scheme_file, f"{extension}:light", "ipynb_to_script"
    )


@pytest.mark.parametrize("extension", ("ss", "scm"))
def test_ipynb_to_scheme_percent(
    ipynb_scheme_file, extension, no_jupytext_version_number
):
    assert_conversion_same_as_mirror(
        ipynb_scheme_file, f"{extension}:percent", "ipynb_to_percent"
    )



import nbformat
import pytest

import jupytext
from jupytext.compare import compare_notebooks


@pytest.mark.parametrize("ext", [".r", ".R"])
def test_identity_source_write_read(ipynb_R_file, ext):
    """
    Test that writing the notebook with R, and read again,
    is the same as removing outputs
    """

    with open(ipynb_R_file) as fp:
        nb1 = nbformat.read(fp, as_version=4)

    R = jupytext.writes(nb1, ext)
    nb2 = jupytext.reads(R, ext)

    compare_notebooks(nb2, nb1)



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [raw] raw_mimetype="text/latex"
# $1+1$

# %% [raw] raw_mimetype="text/restructuredtext"
# :math:`1+1`

# %% [raw] raw_mimetype="text/html"
# <b>Bold text<b>

# %% [raw] raw_mimetype="text/markdown"
# **Bold text**

# %% [raw] raw_mimetype="text/x-python"
# 1 + 1

# %% [raw]
# Not formatted



import pytest
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook

import jupytext
from jupytext.compare import compare, compare_notebooks


@pytest.mark.parametrize("blank_lines", range(1, 6))
def test_file_with_blank_lines(blank_lines):
    py_script = """# Markdown cell{0}
# Another one{0}
""".format(
        "\n".join([""] * blank_lines)
    )

    notebook = jupytext.reads(py_script, "py")
    py_script2 = jupytext.writes(notebook, "py")
    compare(py_script2, py_script)


@pytest.mark.parametrize("blank_cells", range(1, 3))
def test_notebook_with_empty_cells(blank_cells):
    notebook = new_notebook(
        cells=[new_markdown_cell("markdown cell one")]
        + [new_code_cell("") for i in range(blank_cells)]
        + [new_markdown_cell("markdown cell two")]
        + [new_code_cell("") for i in range(blank_cells)],
        metadata={"jupytext": {"main_language": "python"}},
    )

    script = jupytext.writes(notebook, "py")
    notebook2 = jupytext.reads(script, "py")

    compare_notebooks(notebook2, notebook)



import pytest
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell

import jupytext
from jupytext.cell_metadata import (
    _IGNORE_CELL_METADATA,
    RMarkdownOptionParsingError,
    is_valid_metadata_key,
    metadata_to_rmd_options,
    metadata_to_text,
    parse_key_equal_value,
    parse_rmd_options,
    rmd_options_to_metadata,
    text_to_metadata,
    try_eval_metadata,
)
from jupytext.combine import combine_inputs_with_outputs
from jupytext.compare import compare, compare_notebooks
from jupytext.metadata_filter import filter_metadata


def r_options_language_metadata():
    yield from [
        ("r", "R", {}),
        (
            'r plot_1, dpi=72, fig.path="fig_path/"',
            "R",
            {"name": "plot_1", "dpi": 72, "fig.path": "fig_path/"},
        ),
        (
            'r plot_1, bool=TRUE, fig.path="fig_path/"',
            "R",
            {"name": "plot_1", "bool": True, "fig.path": "fig_path/"},
        ),
        ("r echo=FALSE", "R", {"tags": ["remove_input"]}),
        ("r plot_1, echo=TRUE", "R", {"name": "plot_1", "echo": True}),
        (
            "python echo=if a==5 then TRUE else FALSE",
            "python",
            {"echo": "#R_CODE#if a==5 then TRUE else FALSE"},
        ),
        (
            'python noname, tags=c("a", "b", "c"), echo={sum(a+c(1,2))>1}',
            "python",
            {
                "name": "noname",
                "tags": ["a", "b", "c"],
                "echo": "#R_CODE#{sum(a+c(1,2))>1}",
            },
        ),
        ('python active="ipynb,py"', "python", {"active": "ipynb,py"}),
        (
            'python include=FALSE, active="Rmd"',
            "python",
            {"active": "Rmd", "tags": ["remove_cell"]},
        ),
        (
            'r chunk_name, include=FALSE, active="Rmd"',
            "R",
            {"name": "chunk_name", "active": "Rmd", "tags": ["remove_cell"]},
        ),
        ('python tags=c("parameters")', "python", {"tags": ["parameters"]}),
    ]


@pytest.mark.parametrize("options,language, metadata", r_options_language_metadata())
def test_parse_rmd_options(options, language, metadata):
    compare(rmd_options_to_metadata(options), (language, metadata))


@pytest.mark.parametrize("options,language, metadata", r_options_language_metadata())
def test_build_options(options, language, metadata):
    compare(metadata_to_rmd_options(*(language, metadata)), options)


@pytest.mark.parametrize("options,language, metadata", r_options_language_metadata())
def test_build_options_random_order(options, language, metadata):
    # Older python has no respect for order...
    # assert to_chunk_options(metadata) == options

    def split_and_strip(opt):
        {o.strip() for o in opt.split(",")}

    assert split_and_strip(
        metadata_to_rmd_options(*(language, metadata))
    ) == split_and_strip(options)


@pytest.mark.parametrize("options", ["a={)", "name, name2", "a=}", "b=]", "c=["])
def test_parsing_error(options):
    with pytest.raises(RMarkdownOptionParsingError):
        parse_rmd_options(options)


def test_ignore_metadata():
    metadata = {"trusted": True, "tags": ["remove_input"]}
    metadata = filter_metadata(metadata, None, _IGNORE_CELL_METADATA)
    assert metadata_to_rmd_options("R", metadata) == "r echo=FALSE"


def test_filter_metadata():
    assert filter_metadata({"scrolled": True}, None, _IGNORE_CELL_METADATA) == {}


def test_try_eval_metadata():
    metadata = {"list": 'list("a",5)', "c": "c(1,2,3)"}
    try_eval_metadata(metadata, "list")
    try_eval_metadata(metadata, "c")
    assert metadata == {"list": ["a", 5], "c": [1, 2, 3]}


def test_language_no_metadata(text="python", value=("python", {})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text.strip()


def test_only_metadata(text='key="value"', value=("", {"key": "value"})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_only_metadata_2(text='key="value"', value=("", {"key": "value"})):
    compare(text_to_metadata(text, allow_title=True), value)
    assert metadata_to_text(*value) == text


def test_no_language(text=".class", value=("", {".class": None})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_language_metadata_no_space(text='python{"a":1}', value=("python", {"a": 1})):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == "python a=1"


def test_title_no_metadata(text="title", value=("title", {})):
    compare(text_to_metadata(text, allow_title=True), value)
    assert metadata_to_text(*value) == text.strip()


def test_simple_metadata(
    text='python string="value" number=1.0 array=["a", "b"]',
    value=("python", {"string": "value", "number": 1.0, "array": ["a", "b"]}),
):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_simple_metadata_with_spaces(
    text='python string = "value" number = 1.0 array = ["a",  "b"]',
    value=("python", {"string": "value", "number": 1.0, "array": ["a", "b"]}),
):
    compare(text_to_metadata(text), value)
    assert (
        metadata_to_text(*value) == 'python string="value" number=1.0 array=["a", "b"]'
    )


def test_title_and_relax_json(
    text="cell title string='value' number=1.0 array=['a', \"b\"]",
    value=("cell title", {"string": "value", "number": 1.0, "array": ["a", "b"]}),
):
    compare(text_to_metadata(text, allow_title=True), value)
    assert (
        metadata_to_text(*value)
        == 'cell title string="value" number=1.0 array=["a", "b"]'
    )


def test_title_and_json_dict(
    text='cell title {"string": "value", "number": 1.0, "array": ["a", "b"]}',
    value=("cell title", {"string": "value", "number": 1.0, "array": ["a", "b"]}),
):
    compare(text_to_metadata(text, allow_title=True), value)
    assert (
        metadata_to_text(*value)
        == 'cell title string="value" number=1.0 array=["a", "b"]'
    )


@pytest.mark.parametrize("allow_title", [True, False])
def test_attribute(allow_title):
    text = ".class"
    value = ("", {".class": None})
    compare(text_to_metadata(text, allow_title), value)
    assert metadata_to_text(*value) == text


def test_language_and_attribute(
    text="python .class", value=("python", {".class": None})
):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_title_and_attribute(
    text="This is my title. .class", value=("This is my title.", {".class": None})
):
    compare(text_to_metadata(text, allow_title=True), value)
    assert metadata_to_text(*value) == text


def test_values_with_equal_signs_inside(
    text='python string="value=5"', value=("python", {"string": "value=5"})
):
    compare(text_to_metadata(text), value)
    assert metadata_to_text(*value) == text


def test_incorrectly_encoded(text="this is an incorrect expression d={{4 b=3"):
    value = text_to_metadata(text, allow_title=True)
    assert metadata_to_text(*value) == text


def test_incorrectly_encoded_json(text='this is an incorrect expression {"d": "}'):
    value = text_to_metadata(text, allow_title=True)
    assert metadata_to_text(*value) == text


def test_parse_key_value():
    assert parse_key_equal_value("key='value' key2=5.0") == {
        "key": "value",
        "key2": 5.0,
    }


def test_parse_key_value_key():
    assert parse_key_equal_value("key='value' key2") == {
        "key": "value",
        "key2": None,
    }


@pytest.mark.parametrize(
    "key", ["ok", "also.ok", "all_right_55", "not,ok", "unexpected,key"]
)
def test_is_valid_metadata_key(key):
    assert is_valid_metadata_key(key) == ("," not in key)


@pytest.fixture()
def notebook_with_unsupported_key_in_metadata(python_notebook):
    nb = python_notebook
    nb.cells.append(new_markdown_cell("text", metadata={"unexpected,key": True}))
    return nb


def test_unsupported_key_in_metadata(
    notebook_with_unsupported_key_in_metadata, fmt_with_cell_metadata
):
    """Notebook metadata that can't be parsed in text notebook cannot go to the text file,
    but it should still remain available in the ipynb file for paired notebooks."""
    with pytest.warns(
        UserWarning,
        match="The following metadata cannot be exported to the text notebook",
    ):
        text = jupytext.writes(
            notebook_with_unsupported_key_in_metadata, fmt_with_cell_metadata
        )
    assert "unexpected" not in text
    nb_text = jupytext.reads(text, fmt=fmt_with_cell_metadata)
    nb = combine_inputs_with_outputs(
        nb_text, notebook_with_unsupported_key_in_metadata, fmt=fmt_with_cell_metadata
    )
    assert nb.cells[-1].metadata == {"unexpected,key": True}


@pytest.fixture()
def notebook_with_collapsed_cell(python_notebook):
    nb = python_notebook
    nb.cells.append(
        new_markdown_cell("## Data", metadata={"jp-MarkdownHeadingCollapsed": True})
    )
    return nb


def test_notebook_with_collapsed_cell(
    notebook_with_collapsed_cell, fmt_with_cell_metadata
):
    text = jupytext.writes(notebook_with_collapsed_cell, fmt_with_cell_metadata)
    assert "MarkdownHeadingCollapsed" in text
    nb = jupytext.reads(text, fmt=fmt_with_cell_metadata)
    compare_notebooks(nb, notebook_with_collapsed_cell, fmt=fmt_with_cell_metadata)


def test_empty_tags_are_not_saved_in_text_notebooks(
    no_jupytext_version_number, python_notebook, fmt="py:percent"
):
    nb = python_notebook
    nb.cells.append(new_code_cell(metadata={"tags": []}))
    text = jupytext.writes(nb, fmt=fmt)
    assert "tags" not in text
    nb_text = jupytext.reads(text, fmt=fmt)
    nb2 = combine_inputs_with_outputs(nb_text, nb, fmt=fmt)
    assert nb2.cells[-1].metadata["tags"] == []



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 2
#     language: python
#     name: python2
# ---

# %% [markdown]
# # A notebook with R cells
#
# This notebook shows the use of R cells to generate plots

# %%
%load_ext rpy2.ipython

# %%
%%R
suppressMessages(require(tidyverse))

# %%
%%R
ggplot(iris, aes(x = Sepal.Length, y = Petal.Length, color=Species)) + geom_point()

# %% [markdown]
# The default plot dimensions are not good for us, so we use the -w and -h parameters in %%R magic to set the plot size

# %%
%%R -w 400 -h 240
ggplot(iris, aes(x = Sepal.Length, y = Petal.Length, color=Species)) + geom_point()



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

1 + 1


# A markdown cell
# And below, the cell for function f has non trivial cell metadata. And the next cell as well.

# + attributes={"classes": [], "id": "", "n": "10"}
def f(x):
    return x


# + attributes={"classes": [], "id": "", "n": "10"}
f(5)
# -

# More text

2 + 2



import pytest
from git.exc import HookExecutionError
from nbformat.v4.nbbase import new_code_cell
from pre_commit.main import main as pre_commit

from jupytext import read, write


@pytest.mark.requires_user_kernel_python3
def test_pre_commit_hook_sync_execute(
    tmpdir,
    cwd_tmpdir,
    tmp_repo,
    jupytext_repo_root,
    jupytext_repo_rev,
    notebook_with_outputs,
):
    """Here we sync the ipynb notebook with a py:percent file and execute it (this is probably not a very
    recommendable hook!)"""
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [--sync, --execute, --show-changes]
    additional_dependencies:
    - nbconvert
"""
    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)

    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks", "-f"])

    # simple notebook with kernel 'user_kernel_python3'
    nb = notebook_with_outputs
    nb.cells = [new_code_cell("3+4")]

    # pair it to a py file and write it to disk
    nb.metadata["jupytext"] = {"formats": "ipynb,py:percent"}
    write(nb, "test.ipynb")

    # try to commit it, should fail because
    # 1. the py version hasn't been added
    # 2. the outputs are missing
    tmp_repo.git.add("test.ipynb")
    with pytest.raises(
        HookExecutionError,
        match="files were modified by this hook",
    ):
        tmp_repo.index.commit("failing")

    # Add the two files
    tmp_repo.git.add("test.ipynb")
    tmp_repo.git.add("test.py")

    # now the commit will succeed
    tmp_repo.index.commit("passing")
    assert "test.ipynb" in tmp_repo.tree()
    assert "test.py" in tmp_repo.tree()

    # the first cell has the expected output
    nb = read("test.ipynb")
    assert nb.cells[0].outputs[0]["data"]["text/plain"] == "7"



"""Jupyter server and lab extension entry points"""

import asyncio

from jupytext.reraise import reraise

try:
    from jupytext.async_contentsmanager import (
        build_async_jupytext_contents_manager_class,
    )
except ImportError as err:
    build_async_jupytext_contents_manager_class = reraise(err)

try:
    from jupytext.sync_contentsmanager import build_sync_jupytext_contents_manager_class
except ImportError as err:
    build_sync_jupytext_contents_manager_class = reraise(err)


def load_jupyter_server_extension(app):  # pragma: no cover
    """Use Jupytext's contents manager"""
    if hasattr(app.contents_manager_class, "formats"):
        app.log.info(
            "[Jupytext Server Extension] NotebookApp.contents_manager_class is "
            "(a subclass of) jupytext.TextFileContentsManager already - OK"
        )
        return

    # The server extension call is too late!
    # The contents manager was set at NotebookApp.init_configurables

    base_class = app.contents_manager_class

    asynchronous = asyncio.iscoroutinefunction(base_class.get)
    app.log.info(
        "[Jupytext Server Extension] Deriving "
        + ("an Async" if asynchronous else "a ")
        + "TextFileContentsManager from "
        + base_class.__name__
    )

    if asyncio.iscoroutinefunction(base_class.get):
        app.contents_manager_class = build_async_jupytext_contents_manager_class(
            base_class
        )
    else:
        app.contents_manager_class = build_sync_jupytext_contents_manager_class(
            base_class
        )

    try:
        # And rerun selected init steps from https://github.com/jupyter/notebook/blob/
        # 132f27306522b32fa667a6b208034cb7a04025c9/notebook/notebookapp.py#L1634-L1638

        # app.init_configurables()
        app.contents_manager = app.contents_manager_class(parent=app, log=app.log)
        app.session_manager.contents_manager = app.contents_manager

        # app.init_components()
        # app.init_webapp()
        app.web_app.settings["contents_manager"] = app.contents_manager

        # app.init_terminals()
        # app.init_signal()

    except Exception:
        app.log.error(
            """[Jupytext Server Extension] An error occurred. Please deactivate the server extension with
    jupyter serverextension disable jupytext
and configure the contents manager manually by adding
    c.NotebookApp.contents_manager_class = "jupytext.TextFileContentsManager"
to your .jupyter/jupyter_notebook_config.py file.
"""
        )
        raise


def _jupyter_labextension_paths():
    return [{"src": "labextension", "dest": "jupyterlab-jupytext"}]



import jupytext
from jupytext.compare import compare


def test_identity_write_read(rmd_file, no_jupytext_version_number):
    """Test that writing the notebook with ipynb, and read again, yields identity"""

    with open(rmd_file) as fp:
        rmd = fp.read()

    nb = jupytext.reads(rmd, "Rmd")
    rmd2 = jupytext.writes(nb, "Rmd")

    compare(rmd2, rmd)


def test_two_blank_lines_as_cell_separator():
    rmd = """Some markdown
text


And a new cell
"""

    nb = jupytext.reads(rmd, "Rmd")
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[1].cell_type == "markdown"
    assert nb.cells[0].source == "Some markdown\ntext"
    assert nb.cells[1].source == "And a new cell"



import pytest
from nbformat.v4.nbbase import new_notebook, new_raw_cell

from jupytext import reads, writes
from jupytext.compare import compare, compare_notebooks
from jupytext.config import load_jupytext_configuration_file
from jupytext.header import _JUPYTER_METADATA_NAMESPACE
from jupytext.myst import dump_yaml_blocks


@pytest.mark.requires_myst
def test_myst_header_is_stable_1247_using_inline_filter(
    md="""---
jupytext:
  formats: md:myst
  notebook_metadata_filter: -jupytext.text_representation.jupytext_version,settings,mystnb
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
kernelspec:
  display_name: Python 3 (ipykernel)
  language: python
  name: python3
mystnb:
  execution_mode: 'off'
settings:
  output_matplotlib_strings: remove
---
""",
):
    nb = reads(md, fmt="md")
    md2 = writes(nb, fmt="md")

    compare(md2, md)


@pytest.mark.requires_myst
def test_myst_header_is_stable_1247_using_config(
    jupytext_toml_content="""notebook_metadata_filter = "-jupytext.text_representation.jupytext_version,settings,mystnb"
""",
    md="""---
jupytext:
  formats: md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
kernelspec:
  display_name: Python 3 (ipykernel)
  language: python
  name: python3
mystnb:
  execution_mode: 'off'
settings:
  output_matplotlib_strings: remove
---
""",
):
    config = load_jupytext_configuration_file("jupytext.toml", jupytext_toml_content)

    nb = reads(md, fmt="md", config=config)
    md2 = writes(nb, fmt="md", config=config)

    compare(md2, md)


@pytest.mark.requires_myst
def test_myst_frontmatter_metadata_combo(no_jupytext_version_number):
    # ---
    # foo: bar
    # jupyter: <- subkeys become notebook metadata
    #   lorem: ipsum
    # ---
    frontmatter = {"foo": "bar"}
    metadata = {"lorem": "ipsum"}
    md = dump_yaml_blocks(
        {**frontmatter, _JUPYTER_METADATA_NAMESPACE: metadata},
        compact=False,
    )
    config = load_jupytext_configuration_file(
        "jupytext.toml",
        "\n".join(
            (
                'notebook_metadata_filter = "all,-jupytext"',
                'root_level_metadata_filter = "-all"',
            )
        ),
    )
    actual_nb = reads(md, fmt="md:myst", config=config)
    expected_nb = new_notebook(
        metadata=metadata,
        cells=[
            new_raw_cell(dump_yaml_blocks(frontmatter, compact=False).strip()),
        ],
    )
    compare_notebooks(actual_nb, expected_nb)
    # {
    #   "cells": [
    #     {
    #        "cell_type": "raw",
    #        "source": "---\nfoo: bar\n---",  <- merge with metadata in frontmatter
    #        ...
    #     }
    #   ],
    #   "metadata": {"foo": "baz"},
    #   ...
    # }
    actual_md = writes(actual_nb, fmt="md:myst", config=config)
    expected_md = md
    compare(actual_md, expected_md)
